# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from collections import UserList
from typing import TYPE_CHECKING, Annotated, Any, Iterable, Iterator, Optional, Union

import geojson
from pystac import ItemCollection
from shapely.geometry import GeometryCollection
from shapely.geometry import mapping as shapely_mapping
from shapely.geometry import shape
from typing_extensions import Doc

from eodag.api.product import EOProduct
from eodag.plugins.crunch.filter_date import FilterDate
from eodag.plugins.crunch.filter_latest_intersect import FilterLatestIntersect
from eodag.plugins.crunch.filter_latest_tpl_name import FilterLatestByName
from eodag.plugins.crunch.filter_overlap import FilterOverlap
from eodag.plugins.crunch.filter_property import FilterProperty
from eodag.utils import STAC_VERSION, _deprecated

if TYPE_CHECKING:
    from shapely.geometry.base import BaseGeometry

    from eodag.api.core import EODataAccessGateway
    from eodag.plugins.crunch.base import Crunch


logger = logging.getLogger("eodag.search_result")


class SearchResult(UserList[EOProduct]):
    """An object representing a collection of :class:`~eodag.api.product._product.EOProduct` resulting from a search.

    :param products: A list of products resulting from a search
    :param number_matched: (optional) the estimated total number of matching results
    :param errors: (optional) stored errors encountered during the search. Tuple of (provider name, exception)
    :param search_params: (optional) search parameters stored to use in pagination
    :param next_page_token: (optional) next page token value to use in pagination
    :param next_page_token_key: (optional) next page token key to use in pagination
    :param raise_errors: (optional) whether to raise errors encountered during the search

    :cvar data: List of products
    :ivar number_matched: Estimated total number of matching results
    """

    errors: Annotated[
        list[tuple[str, Exception]], Doc("Tuple of provider name, exception")
    ]

    def __init__(
        self,
        products: list[EOProduct],
        number_matched: Optional[int] = None,
        errors: Optional[list[tuple[str, Exception]]] = None,
        search_params: Optional[dict[str, Any]] = None,
        next_page_token: Optional[str] = None,
        next_page_token_key: Optional[str] = None,
        raise_errors: Optional[bool] = False,
    ) -> None:
        super().__init__(products)
        self.number_matched = number_matched
        self.errors = errors if errors is not None else []
        self.search_params = search_params
        self.next_page_token = next_page_token
        self.next_page_token_key = next_page_token_key
        self.raise_errors = raise_errors
        self._dag: Optional["EODataAccessGateway"] = None

    def crunch(self, cruncher: Crunch, **search_params: Any) -> SearchResult:
        """Do some crunching with the underlying EO products.

        :param cruncher: The plugin instance to use to work on the products
        :param search_params: The criteria that have been used to produce this result
        :returns: The result of the application of the crunching method to the EO products
        """
        crunched_results = cruncher.proceed(self.data, **search_params)
        return SearchResult(crunched_results)

    def filter_date(
        self, start: Optional[str] = None, end: Optional[str] = None
    ) -> SearchResult:
        """Filter products by date.

        Allows to filter out products that are older than a start date (optional) or more recent than an end date
        (optional).

        Applies :class:`~eodag.plugins.crunch.filter_date.FilterDate` crunch.

        :param start: start sensing time in iso format
        :param end: end sensing time in iso format
        :returns: The result of the application of the crunching method to the EO products
        """
        return self.crunch(FilterDate(dict(start=start, end=end)))

    def filter_latest_intersect(
        self, geometry: Union[dict[str, Any], BaseGeometry, Any]
    ) -> SearchResult:
        """Filter latest products (the ones with a the highest start date) that intersect search extent.

        Applies :class:`~eodag.plugins.crunch.filter_latest_intersect.FilterLatestIntersect` crunch.

        :param geometry: geometry used as search extent.
        :returns: The result of the application of the crunching method to the EO products
        """
        return self.crunch(FilterLatestIntersect({}), geometry=geometry)

    def filter_latest_by_name(self, name_pattern: str) -> SearchResult:
        """Filter Search results to get only the latest product, based on the name of the product.

        Applies :class:`~eodag.plugins.crunch.filter_latest_tpl_name.FilterLatestByName` crunch.

        :param name_pattern: 6 digits product name pattern (tile id)
        :returns: The result of the application of the crunching method to the EO products
        """
        return self.crunch(FilterLatestByName(dict(name_pattern=name_pattern)))

    def filter_overlap(
        self,
        geometry: Any,
        minimum_overlap: int = 0,
        contains: bool = False,
        intersects: bool = False,
        within: bool = False,
    ) -> SearchResult:
        """Filter products, retaining only those that are overlapping with the search_extent.

        Applies :class:`~eodag.plugins.crunch.filter_overlap.FilterOverlap` crunch.

        :param geometry: geometry used as search extent
        :param minimum_overlap: minimal overlap percentage
        :param contains: ``True`` if product geometry contains the search area
        :param intersects: ``True`` if product geometry intersects the search area
        :param within: ``True`` if product geometry is within the search area
        :returns: The result of the application of the crunching method to the EO products
        """
        return self.crunch(
            FilterOverlap(
                dict(
                    minimum_overlap=minimum_overlap,
                    contains=contains,
                    intersects=intersects,
                    within=within,
                )
            ),
            geometry=geometry,
        )

    def filter_property(
        self, operator: str = "eq", **search_property: Any
    ) -> SearchResult:
        """Filter products, retaining only those whose property match criteria.

        Applies :class:`~eodag.plugins.crunch.filter_property.FilterProperty` crunch.

        :param operator: Operator used for filtering (one of :mod:`python:operator` functions ``lt,le,eq,ne,ge,...``)
        :param search_property: property key from ``product.properties``, associated to its filter value
        """
        return self.crunch(FilterProperty(dict(operator=operator, **search_property)))

    def filter_online(self) -> SearchResult:
        """Filter to only keep online products.

        Applies :class:`~eodag.plugins.crunch.filter_property.FilterProperty` crunch for ``order:status == succeeded``.

        :returns: The result of the application of the crunching method to the EO products
        """
        return self.filter_property(**{"order:status": "succeeded"})

    @classmethod
    def from_dict(
        cls,
        feature_collection: dict[str, Any],
        dag: Optional[EODataAccessGateway] = None,
    ) -> SearchResult:
        """Builds an :class:`~eodag.api.search_result.SearchResult` object from its representation as geojson

        :param feature_collection: A collection representing a search result.
        :param dag: (optional) The EODataAccessGateway instance to use for registering the products.
        :returns: An eodag representation of a search result
        """

        products = [
            EOProduct.from_dict(feature, dag=dag)
            for feature in feature_collection.get("features", [])
        ]
        props = feature_collection.get("metadata", {}) or {}

        eodag_search_params = props.get("eodag:search_params", {})
        if eodag_search_params and eodag_search_params.get("geometry"):
            eodag_search_params["geometry"] = shape(eodag_search_params["geometry"])

        results = cls(
            products=products,
            number_matched=props.get("eodag:number_matched"),
            next_page_token=props.get("eodag:next_page_token"),
            next_page_token_key=props.get("eodag:next_page_token_key"),
            search_params=eodag_search_params or None,
            raise_errors=props.get("eodag:raise_errors"),
        )
        if dag is not None:
            results._dag = dag
        return results

    @classmethod
    def from_file(
        cls,
        filepath: str,
        dag: Optional[EODataAccessGateway] = None,
    ) -> SearchResult:
        """Builds an :class:`~eodag.api.search_result.SearchResult` object from a geojson file

        :param filepath: Path to the file containing the feature collection.
        :param dag: (optional) The EODataAccessGateway instance to use for registering the products.
        :returns: An eodag representation of a search result
        """
        with open(filepath, "r") as fh:
            feature = geojson.load(fh)

        return cls.from_dict(feature, dag=dag)

    @classmethod
    def from_pystac(
        cls,
        item_collection: ItemCollection,
        dag: Optional[EODataAccessGateway] = None,
    ) -> SearchResult:
        """Builds an :class:`~eodag.api.search_result.SearchResult` object from a pystac ItemCollection

        :param item_collection: The :class:`pystac.ItemCollection` containing the metadata of the products.
        :param dag: (optional) The EODataAccessGateway instance to use for registering the products.
        :returns: An eodag representation of a search result
        """
        features_collection = item_collection.to_dict()

        return cls.from_dict(features_collection, dag=dag)

    @staticmethod
    @_deprecated(
        reason="Please use 'SearchResult.from_dict' instead",
        version="4.1.0",
    )
    def from_geojson(feature_collection: dict[str, Any]) -> SearchResult:
        """Builds an :class:`~eodag.api.search_result.SearchResult` object from its representation as geojson

        :param feature_collection: A collection representing a search result.
        :returns: An eodag representation of a search result
        """
        return SearchResult.from_dict(feature_collection)

    def as_dict(self, skip_invalid: bool = True) -> dict[str, Any]:
        """GeoJSON representation of SearchResult

        :param skip_invalid: Whether to skip properties whose values are not valid according to the STAC specification.
        :returns: The representation of a :class:`~eodag.api.search_result.SearchResult` as a Python dict
        """

        geojson_search_params = {} | (self.search_params or {})
        # search_params shapely geometry to wkt
        if self.search_params and self.search_params.get("geometry"):
            geojson_search_params["geometry"] = shapely_mapping(
                self.search_params["geometry"]
            )

        return {
            "type": "FeatureCollection",
            "features": [
                product.as_dict(skip_invalid=skip_invalid) for product in self
            ],
            "metadata": {
                "eodag:number_matched": self.number_matched,
                "eodag:next_page_token": self.next_page_token,
                "eodag:next_page_token_key": self.next_page_token_key,
                "eodag:search_params": geojson_search_params or None,
                "eodag:raise_errors": self.raise_errors,
            },
            "links": [],
            "stac_extensions": [],
            "stac_version": STAC_VERSION,
        }

    @_deprecated(
        reason="Please use 'SearchResult.as_dict' instead",
        version="4.1.0",
    )
    def as_geojson_object(self, skip_invalid: bool = True) -> dict[str, Any]:
        """GeoJSON representation of SearchResult

        :param skip_invalid: Whether to skip properties whose values are not valid according to the STAC specification.
        :returns: The representation of a :class:`~eodag.api.search_result.SearchResult` as a Python dict
        """
        return self.as_dict(skip_invalid=skip_invalid)

    def as_shapely_geometry_object(
        self, skip_invalid: bool = True
    ) -> GeometryCollection:
        """:class:`shapely.GeometryCollection` representation of SearchResult

        :param skip_invalid: Whether to skip properties whose values are not valid according to the STAC specification.
        :returns: The representation of a :class:`~eodag.api.search_result.SearchResult` as a
                  :class:`shapely.GeometryCollection`
        """
        return GeometryCollection(
            [
                shape(feature["geometry"]).buffer(0)
                for feature in self.as_dict(skip_invalid=skip_invalid)["features"]
            ]
        )

    def as_wkt_object(self, skip_invalid: bool = True) -> str:
        """WKT representation of SearchResult

        :param skip_invalid: Whether to skip properties whose values are not valid according to the STAC specification.
        :returns: The representation of a :class:`~eodag.api.search_result.SearchResult` as a WKT string
        """
        return self.as_shapely_geometry_object(skip_invalid=skip_invalid).wkt

    def as_pystac_object(self, skip_invalid: bool = True) -> ItemCollection:
        """Pystac ItemCollection representation of SearchResult

        :param skip_invalid: Whether to skip properties whose values are not valid according to the STAC specification.
        :returns: The representation of a :class:`~eodag.api.search_result.SearchResult` as a
                  :class:`pystac.ItemCollection`
        """
        results_dict = self.as_dict(skip_invalid=skip_invalid)
        return ItemCollection.from_dict(results_dict)

    @property
    def __geo_interface__(self) -> dict[str, Any]:
        """Implements the geo-interface protocol.

        See https://gist.github.com/sgillies/2217756
        """
        return self.as_geojson_object()

    def _repr_html_(self):
        total_count = f"/{self.number_matched}" if self.number_matched else ""
        return (
            f"""<table>
                <thead><tr><td style='text-align: left; color: grey;'>
                 {type(self).__name__}&ensp;({len(self)}{total_count})
                </td></tr></thead>
            """
            + "".join(
                [
                    f"""<tr><td style='text-align: left;'>
                <details><summary style='color: grey; font-family: monospace;'>
                    {i}&ensp;
                    {type(p).__name__}(id=<span style='color: black;'>{
                        p.properties["id"]
                    }</span>, provider={p.provider})
                </summary>
                {p._repr_html_()}
                </details>
                </td></tr>
                """
                    for i, p in enumerate(self)
                ]
            )
            + "</table>"
        )

    def extend(self, other: Iterable) -> None:
        """override extend method to include errors"""
        if isinstance(other, SearchResult):
            self.errors.extend(other.errors)

        return super().extend(other)

    def next_page(self, update: bool = True) -> Iterator[SearchResult]:
        """
        Retrieve and iterate over the next pages of search results, if available.

        This method uses the current search parameters and next page token to request
        additional results from the provider. If ``update`` is ``True``, the current ``SearchResult``
        instance is updated with new products and pagination information as pages are fetched.

        :param update: If ``True``, update the current ``SearchResult`` with new results.
        :returns: An iterator yielding ``SearchResult`` objects for each subsequent page.

        Example:

        >>> first_page = SearchResult([])  # result of a search
        >>> for new_results in first_page.next_page():
        ...     continue  # do something with new_results
        """

        def get_next_page(current):
            if current.search_params is None:
                current.search_params = {}
            # Update the next_page_token in the search params
            current.search_params["next_page_token"] = current.next_page_token
            current.search_params["next_page_token_key"] = current.next_page_token_key
            # Ensure the provider is in the search params
            if "provider" in current.search_params:
                current.search_params["provider"] = current.search_params.get(
                    "provider", None
                )
            elif current.data and hasattr(current.data[-1], "provider"):
                current.search_params["provider"] = current.data[-1].provider
            search_plugins, search_kwargs = current._dag._prepare_search(
                **current.search_params
            )
            # If number_matched was provided, ensure it is passed to the next search
            if current.number_matched:
                search_kwargs["number_matched"] = current.number_matched
            for i, search_plugin in enumerate(search_plugins):
                # validate no needed for next pages
                search_kwargs["validate"] = False
                return current._dag._do_search(
                    search_plugin,
                    raise_errors=self.raise_errors,
                    **search_kwargs,
                )

        # Do not iterate if there is no next page token
        #  or if the current one returned less than the maximum number of items asked for.
        if self.next_page_token is None:
            return

        new_results = get_next_page(self)
        old_results = self

        while new_results.data:
            # The products between two iterations are compared. If they
            # are actually the same product, it means the iteration failed at
            # progressing for some reason.
            if (
                old_results.data[0].properties["id"]
                == new_results.data[0].properties["id"]
            ):
                logger.warning(
                    "Iterate over pages: stop iterating since the next page "
                    "appears to have the same products as in the previous one. "
                    "This provider may not implement pagination.",
                )
                break
            if update:
                self.data += new_results.data
                self.search_params = new_results.search_params
                self.next_page_token = new_results.next_page_token
                self.next_page_token_key = new_results.next_page_token_key
            yield new_results
            # Stop iterating if there is no next page token
            #  or if the current one returned less than the maximum number of items asked for.
            if (
                new_results.next_page_token is None
                or len(new_results) < new_results.search_params["limit"]
            ):
                break
            old_results = new_results
            new_results = get_next_page(new_results)
            if not new_results:
                break
        return


class RawSearchResult(UserList[dict[str, Any]]):
    """An object representing a collection of raw/unparsed search results obtained from a provider.

    :param results: A list of raw/unparsed search results
    """

    query_params: dict[str, Any]
    collection_def_params: dict[str, Any]
    search_params: dict[str, Any]
    next_page_token: Optional[str] = None
    next_page_token_key: Optional[str] = None

    def __init__(self, results: list[Any]) -> None:
        super(RawSearchResult, self).__init__(results)



# -*- coding: utf-8 -*-
# Copyright 2024, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import copy
import datetime as dt
import logging
import os
import re
from typing import TYPE_CHECKING, Any, Optional, cast
from urllib.parse import urlsplit

import boto3
import botocore
import requests
from dateutil.tz import tzutc
from dateutil.utils import today

from eodag import EOProduct
from eodag.api.product import AssetsDict
from eodag.api.search_result import SearchResult
from eodag.config import PluginConfig
from eodag.plugins.search import PreparedSearch
from eodag.plugins.search.static_stac_search import StaticStacSearch
from eodag.utils import get_bucket_name_and_prefix, get_geometry_from_various
from eodag.utils.dates import parse_to_utc, to_iso_utc_string
from eodag.utils.exceptions import RequestError, UnsupportedCollection, ValidationError

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client

logger = logging.getLogger("eodag.search.cop_marine")


def _get_date_from_yyyymmdd(date_str: str, item_key: str) -> Optional[dt.datetime]:
    year = date_str[:4]
    month = date_str[4:6]
    if len(date_str) > 6:
        day = date_str[6:]
    else:
        day = "1"
    try:
        date = dt.datetime(
            int(year),
            int(month),
            int(day),
            tzinfo=tzutc(),
        )
    except ValueError:
        logger.error(f"{item_key}: {date_str} is not a valid date")
        return None
    else:
        return date


def _get_dates_from_dataset_data(
    dataset_item: dict[str, Any]
) -> Optional[dict[str, str]]:
    dates = {}
    if "start_datetime" in dataset_item["properties"]:
        dates["start"] = dataset_item["properties"]["start_datetime"]
        dates["end"] = dataset_item["properties"]["end_datetime"]
    elif "datetime" in dataset_item["properties"]:
        dates["start"] = dataset_item["properties"]["datetime"]
        dates["end"] = dataset_item["properties"]["datetime"]
    else:
        return None
    return dates


def _get_s3_client(endpoint_url: str) -> S3Client:
    s3_session = boto3.Session()
    return s3_session.client(
        "s3",
        config=botocore.config.Config(
            # Configures to use subdomain/virtual calling format.
            s3={"addressing_style": "virtual"},
            signature_version=botocore.UNSIGNED,
        ),
        endpoint_url=endpoint_url,
    )


def _check_int_values_properties(properties: dict[str, Any]):
    # remove int values with a bit length of more than 64 from the properties
    invalid = []
    for prop, prop_value in properties.items():
        if isinstance(prop_value, int) and prop_value.bit_length() > 64:
            invalid.append(prop)
        if isinstance(prop_value, dict):
            _check_int_values_properties(prop_value)

    for inv_key in invalid:
        properties.pop(inv_key)


class CopMarineSearch(StaticStacSearch):
    """class that implements search for the Copernicus Marine provider

    It calls :meth:`~eodag.plugins.search.static_stac_search.StaticStacSearch.discover_collections`
    inherited from :class:`~eodag.plugins.search.static_stac_search.StaticStacSearch`
    but for the actual search a special method which fetches the urls of the available products from an S3 storage and
    filters them has been written.

    The configuration parameters are inherited from the parent and grand-parent classes. The
    :attr:`~eodag.config.PluginConfig.DiscoverMetadata.auto_discovery` parameter in the
    :attr:`~eodag.config.PluginConfig.discover_metadata` section has to be set to ``false`` and the
    :attr:`~eodag.config.PluginConfig.DiscoverQueryables.fetch_url` in the
    :attr:`~eodag.config.PluginConfig.discover_queryables` queryables section has to be set to ``null`` to
    overwrite the default config from the stac provider configuration because those functionalities
    are not available.
    """

    def __init__(self, provider: str, config: PluginConfig):
        original_metadata_mapping = copy.deepcopy(config.metadata_mapping)
        super().__init__(provider, config)
        # reset to original metadata mapping from config (changed in super class init)
        self.config.metadata_mapping = original_metadata_mapping

    def _get_collection_info(
        self, collection: str
    ) -> tuple[dict[str, Any], list[dict[str, Any]]]:
        """Fetch collection and associated datasets info"""

        fetch_url = cast(str, self.config.discover_collections["fetch_url"]).format(
            **self.config.__dict__
        )

        logger.debug("fetch data for collection %s", collection)
        provider_collection = self.config.products.get(collection, {}).get(
            "_collection"
        )
        if not provider_collection:
            provider_collection = collection
        collection_url = (
            fetch_url.replace("catalog.stac.json", provider_collection)
            + "/product.stac.json"
        )
        try:
            collection_data = requests.get(collection_url).json()
        except requests.RequestException as exc:
            if exc.errno == 404:
                logger.error("product %s not found", collection)
                raise UnsupportedCollection(collection)
            logger.error("data for product %s could not be fetched", collection)
            raise RequestError.from_error(
                exc, f"data for product {collection} could not be fetched"
            ) from exc

        datasets = []
        for link in [li for li in collection_data["links"] if li["rel"] == "item"]:
            dataset_url = (
                fetch_url.replace("catalog.stac.json", provider_collection)
                + "/"
                + link["href"]
            )
            try:
                dataset_item = requests.get(dataset_url).json()
                datasets.append(dataset_item)
            except requests.RequestException:
                logger.error("data for dataset %s could not be fetched", link["title"])

        return collection_data, datasets

    def _create_product(
        self,
        collection: str,
        item_key: str,
        s3_url: str,
        dataset_item: dict[str, Any],
        collection_dict: dict[str, Any],
        use_dataset_dates: bool = False,
        product_id: Optional[str] = None,
        asset_properties: dict = {},
    ) -> Optional[EOProduct]:

        item_id = os.path.splitext(item_key.split("/")[-1])[0]
        if product_id and product_id != item_id:
            return None
        s3_url_parts = s3_url.split("/")
        # create download url from s3_url and item_key,
        # there are situations where the 2 overlap -> remove parts included in s3_url from item_key
        item_key_parts = [
            part for part in item_key.split("/") if part not in s3_url_parts
        ]
        download_url = s3_url + "/" + "/".join(item_key_parts)
        geometry = (
            get_geometry_from_various(**dataset_item)
            or self.config.metadata_mapping["eodag:default_geometry"]
        )
        properties = {
            "id": item_id,
            "title": item_id,
            "geometry": geometry,
            "eodag:download_link": download_url,
            "dataset": dataset_item["id"],
            # order:status set to succeeded for consistency between providers
            "order:status": "succeeded",
        }
        if use_dataset_dates:
            dates = _get_dates_from_dataset_data(dataset_item)
            if not dates:
                return None
            properties["start_datetime"] = dates["start"]
            properties["end_datetime"] = dates["end"]
        else:
            item_dates = re.findall(r"(\d{4})(0[1-9]|1[0-2])([0-3]\d)", item_id)
            if not item_dates:
                item_dates = re.findall(r"_(\d{4})(0[1-9]|1[0-2])", item_id)
            item_dates = ["".join(row) for row in item_dates]
            item_start = _get_date_from_yyyymmdd(item_dates[0], item_key)
            if not item_start:  # identified pattern was not a valid datetime
                return None
            if len(item_dates) > 2:  # start, end and created_at timestamps
                item_end = _get_date_from_yyyymmdd(item_dates[1], item_key)
            else:  # only date and created_at timestamps
                item_end = item_start
            properties["start_datetime"] = to_iso_utc_string(item_start)
            properties["end_datetime"] = to_iso_utc_string(item_end or item_start)

        for key, value in collection_dict["properties"].items():
            if key not in ["id", "title", "start_datetime", "end_datetime", "datetime"]:
                properties[key] = value
        for key, value in dataset_item["properties"].items():
            if key not in ["id", "title", "start_datetime", "end_datetime", "datetime"]:
                properties[key] = value

        code_mapping = self.config.products.get(collection, {}).get("code_mapping")
        if code_mapping:
            id_parts = item_id.split("_")
            if len(id_parts) > code_mapping["index"]:
                code = id_parts[code_mapping["index"]]
                if "pattern" not in code_mapping:
                    properties[code_mapping["param"]] = code
                elif re.findall(code_mapping["pattern"], code):
                    properties[code_mapping["param"]] = re.findall(
                        code_mapping["pattern"], code
                    )[0]

        _check_int_values_properties(properties)

        properties["eodag:thumbnail"] = collection_dict["assets"]["thumbnail"]["href"]
        if "omiFigure" in collection_dict["assets"]:
            properties["eodag:quicklook"] = collection_dict["assets"]["omiFigure"][
                "href"
            ]

        asset_native = {
            "title": "native",
            "href": download_url,
            "type": "application/x-netcdf",
        }
        asset_native.update(asset_properties)
        assets = {"native": asset_native}
        additional_assets = self.get_assets_from_mapping(dataset_item)
        assets.update(additional_assets)

        product = EOProduct(self.provider, properties, collection=collection)
        product.assets = AssetsDict(product, assets)
        product._normalize_bands()
        return product

    def query(
        self,
        prep: PreparedSearch = PreparedSearch(),
        **kwargs: Any,
    ) -> SearchResult:
        """
        Implementation of search for the Copernicus Marine provider
        :param prep: object containing search parameterds
        :param kwargs: additional search arguments
        :returns: list of products and total number of products
        """
        limit = prep.limit
        token_value = getattr(prep, "next_page_token") or prep.page

        # only return 1 page if pagination is disabled
        if token_value is None or limit is None or int(token_value) > 1 and limit <= 0:
            result = SearchResult([])
            if prep.count:
                result.number_matched = 0
            return result

        token = int(token_value)

        collection = kwargs.get("collection", prep.collection)
        if not collection:
            raise ValidationError(
                "parameter collection is required for search with cop_marine provider"
            )
        collection_dict, datasets_items_list = self._get_collection_info(collection)
        geometry = kwargs.pop("geometry", None)
        products: list[EOProduct] = []
        start_index = limit * (token - 1) + 1
        num_total = 0
        for i, dataset_item in enumerate(datasets_items_list):
            if len(products) >= limit and not prep.count and limit > 0:
                break
            # Filter by geometry
            if "id" not in kwargs and geometry:
                dataset_geom = get_geometry_from_various(**dataset_item)
                if dataset_geom and not dataset_geom.intersects(geometry):
                    continue
            try:
                logger.debug("searching data for dataset %s", dataset_item["id"])

                # date bounds
                if "start_datetime" in kwargs:
                    start_date = parse_to_utc(kwargs["start_datetime"])
                elif "start_datetime" in dataset_item["properties"]:
                    start_date = parse_to_utc(
                        dataset_item["properties"]["start_datetime"]
                    )
                else:
                    start_date = parse_to_utc(dataset_item["properties"]["datetime"])
                if "end_datetime" in kwargs:
                    end_date = parse_to_utc(kwargs["end_datetime"])
                elif "end_datetime" in dataset_item["properties"]:
                    end_date = parse_to_utc(dataset_item["properties"]["end_datetime"])
                else:
                    end_date = today(tzinfo=tzutc())

                # retrieve information about s3 from collection data
                s3_url = dataset_item["assets"]["native"]["href"]
            except KeyError as e:
                logger.warning(
                    f"Unable to extract info from {collection} item #{i}: {str(e)}"
                )
                continue

            url_parts = urlsplit(s3_url)
            endpoint_url = url_parts.scheme + "://" + url_parts.hostname
            bucket, collection_path = get_bucket_name_and_prefix(s3_url, 0)
            if bucket is None or collection_path is None:
                logger.warning(
                    f"Unable to get bucket and prefix from {s3_url}, got {(bucket, collection_path)}"
                )
                continue

            s3_client = _get_s3_client(endpoint_url)
            asset_properties: dict[str, Any] = {}

            if ".nc" in collection_path:
                num_total += 1
                if num_total < start_index:
                    continue

                if len(products) < limit or limit < 0:
                    asset_properties = {}
                    try:
                        object_metadata = s3_client.head_object(
                            Bucket=bucket, Key=collection_path
                        )
                        if (
                            object_metadata.get("ResponseMetadata", {}).get(
                                "HTTPStatusCode"
                            )
                            == 200
                        ):
                            headers = object_metadata.get("ResponseMetadata", {}).get(
                                "HTTPHeaders", {}
                            )
                            # do not use 'content-type' header, always at value binary/octet-stream
                            if (
                                content_length := headers.get("content-length")
                            ) is not None:
                                asset_properties["file:size"] = int(content_length)
                            if (etag := headers.get("etag")) is not None:
                                if "-" not in etag:
                                    asset_properties["file:checksum"] = etag.strip('"')
                            if (
                                last_modified := headers.get("last-modified")
                            ) is not None:
                                try:
                                    asset_properties["updated"] = to_iso_utc_string(
                                        last_modified
                                    )
                                except Exception:
                                    pass
                    except Exception:
                        pass

                    product = self._create_product(
                        collection,
                        collection_path,
                        endpoint_url + "/" + bucket,
                        dataset_item,
                        collection_dict,
                        True,
                        kwargs.get("id"),
                        asset_properties=asset_properties,
                    )
                    if product:
                        products.append(product)
                    if product and kwargs.get("id"):
                        break
                    continue

            stop_search = False
            current_object = None
            while not stop_search:
                # list_objects returns max 1000 objects -> use marker to get next objects
                if current_object:
                    s3_objects = s3_client.list_objects(
                        Bucket=bucket, Prefix=collection_path, Marker=current_object
                    )
                else:
                    s3_objects = s3_client.list_objects(
                        Bucket=bucket, Prefix=collection_path
                    )
                if "Contents" not in s3_objects:
                    if len(products) == 0 and i == len(datasets_items_list) - 1:
                        result = SearchResult([])
                        if prep.count:
                            result.number_matched = 0
                        return result
                    else:
                        break
                elif len(s3_objects["Contents"]) == 0:
                    stop_search = True
                    break

                for obj in s3_objects["Contents"]:
                    item_key = obj["Key"]
                    item_id = os.path.splitext(item_key.split("/")[-1])[0]
                    # filter according to date(s) in item id
                    item_dates = re.findall(r"(\d{4})(0[1-9]|1[0-2])([0-3]\d)", item_id)
                    if not item_dates:
                        item_dates = re.findall(r"_(\d{4})(0[1-9]|1[0-2])", item_id)
                    item_dates = [
                        "".join(row) for row in item_dates
                    ]  # join tuples returned by findall

                    if "id" in kwargs:
                        if kwargs["id"] in obj["Key"]:
                            product = self._create_product(
                                collection,
                                obj["Key"],
                                s3_url,
                                dataset_item,
                                collection_dict,
                                not bool(item_dates),
                            )
                            if product:
                                return SearchResult([product], 1)
                        if len(s3_objects["Contents"]) > 0:
                            current_object = s3_objects["Contents"][-1]["Key"]
                            continue

                    item_start = None
                    item_end = None
                    use_dataset_dates = False
                    if item_dates:
                        item_start = _get_date_from_yyyymmdd(item_dates[0], item_key)
                        if len(item_dates) > 2:  # start, end and created_at timestamps
                            item_end = _get_date_from_yyyymmdd(item_dates[1], item_key)
                    if not item_start:
                        # no valid datetime given in id
                        use_dataset_dates = True
                        dates = _get_dates_from_dataset_data(dataset_item)
                        if dates:
                            item_start = parse_to_utc(dates["start"])
                            item_end = parse_to_utc(dates["end"])
                    if not item_start:
                        # no valid datetime in id and dataset data
                        continue
                    if item_start > end_date:
                        stop_search = True
                    if (
                        (start_date <= item_start <= end_date)
                        or (item_end and start_date <= item_end <= end_date)
                        or (
                            item_end and item_start < start_date and item_end > end_date
                        )
                    ):
                        num_total += 1
                        if num_total < start_index:
                            continue

                        if len(products) < limit or limit < 0:

                            # Asset properties
                            asset_properties = {}

                            last_modified_date = obj.get("LastModified")
                            if isinstance(last_modified_date, dt.datetime):
                                asset_properties["updated"] = to_iso_utc_string(
                                    last_modified_date
                                )

                            etag_obj: Any = obj.get("ETag")
                            if isinstance(etag_obj, str) and "-" not in etag_obj:
                                asset_properties["file:checksum"] = etag_obj.strip('"')

                            size = obj.get("Size")
                            if size is not None:
                                asset_properties["file:size"] = int(size)

                            owner_id = obj.get("Owner", {}).get("ID")
                            if owner_id is not None:
                                asset_properties["cop_marine:owner_id"] = owner_id

                            owner_displayname = obj.get("Owner", {}).get("DisplayName")
                            if owner_displayname is not None:
                                asset_properties[
                                    "cop_marine:owner_name"
                                ] = owner_displayname

                            product = self._create_product(
                                collection,
                                item_key,
                                endpoint_url + "/" + bucket,
                                dataset_item,
                                collection_dict,
                                use_dataset_dates,
                                asset_properties=asset_properties,
                            )
                            if product:
                                products.append(product)
                    current_object = item_key
                    if len(products) >= limit and not prep.count:
                        stop_search = True
                        break

        search_params = (
            kwargs
            | {"limit": prep.limit}
            | {"collection": collection}
            | {"provider": self.provider}
            | {"geometry": geometry}
            if geometry
            else {}
        )

        number_matched = num_total if prep.count else None

        formated_result = SearchResult(
            products,
            number_matched,
            search_params=search_params,
            next_page_token=str(start_index + 1),
        )
        return formated_result



import os
import signal
import threading
from dataclasses import dataclass, field
from typing import Iterable, Iterator, Mapping, Optional, Union


class StreamResponseContent(Iterable[bytes]):
    """
    ByteIO minimal compatibility, used by boto3.upload_fileobj that's
    usually expect BytesIO object, not Iterable[bytes]
    """

    __initialized: bool = False
    __instances: list["StreamResponseContent"] = []

    @staticmethod
    def install_signal_handlers() -> bool:
        """Register SIGINT/SIGTERM handlers that interrupt any live stream.

        This must be called explicitly from the main thread (typically during a
        server's startup) because :func:`signal.signal` only works there. It is
        a no-op if the handlers are already installed or if it is called outside
        of the main thread.

        :returns: ``True`` if the handlers were installed, ``False`` otherwise.
        """
        if StreamResponseContent.__initialized:
            return False
        if threading.current_thread() is not threading.main_thread():
            return False

        StreamResponseContent.__initialized = True

        # Catch end of main process to internal status
        def signal_handler(sig, frame):
            for stream in StreamResponseContent.__instances:
                stream.interrupt()
            StreamResponseContent.__instances = []

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        return True

    def __init__(self, content: Union[Iterable[bytes], bytes]):
        self.buffer: bytes = b""
        self.interrupted: bool = False
        StreamResponseContent.__instances.append(self)
        if isinstance(content, bytes):
            content = [content]
        self.iterator: Iterator[bytes] = iter(content)

    def __iter__(self) -> Iterator[bytes]:
        return self.iterator

    def interrupt(self):
        if not self.interrupted:
            self.interrupted = True

    def read(self, size=-1) -> bytes:

        # Fill the buffer with enough bytes to fulfil the request.
        while not self.interrupted and (size < 0 or len(self.buffer) < size):
            try:
                chunk = next(self.iterator)
                self.buffer = self.buffer + bytes(chunk)
            except StopIteration:
                break

        if self.interrupted:
            raise InterruptedError()

        result: bytes = b""
        if size < 0:
            result, self.buffer = self.buffer, b""
        else:
            result, self.buffer = self.buffer[:size], self.buffer[size:]

        return bytes(result)


@dataclass
class StreamResponse:
    """Represents a streaming response"""

    content: StreamResponseContent
    _filename: Optional[str] = field(default=None, repr=False, init=False)
    _size: Optional[int] = field(default=None, repr=False, init=False)
    headers: dict[str, str] = field(default_factory=dict)
    media_type: Optional[str] = None
    status_code: Optional[int] = None
    arcname: Optional[str] = None

    def __init__(
        self,
        content: Union[Iterable[bytes], bytes],
        filename: Optional[str] = None,
        size: Optional[int] = None,
        headers: Optional[Mapping[str, str]] = None,
        media_type: Optional[str] = None,
        status_code: Optional[int] = None,
        arcname: Optional[str] = None,
    ):
        self.content = StreamResponseContent(content)
        self.headers = dict(headers) if headers else {}
        self.media_type = media_type
        self.status_code = status_code
        self.arcname = arcname
        # use property setters to update headers
        self.filename = filename
        self.size = size

    # filename handling
    @property
    def filename(self) -> Optional[str]:
        """Get the filename for the streaming response.

        :returns: The filename, or None if not set
        """
        return self._filename

    @filename.setter
    def filename(self, value: Optional[str]) -> None:
        """Set the filename and update the Content-Disposition header accordingly.

        :param value: The filename to set, or None to clear it
        """
        self._filename = value
        if value:
            outputs_filename = os.path.basename(value)
            self.headers[
                "Content-Disposition"
            ] = f'attachment; filename="{outputs_filename}"'
        elif "Content-Disposition" in self.headers:
            del self.headers["Content-Disposition"]

    # size handling
    @property
    def size(self) -> Optional[int]:
        """Get the content size for the streaming response.

        :returns: The content size in bytes, or None if not set
        """
        return self._size

    @size.setter
    def size(self, value: Optional[int]) -> None:
        """Set the content size and update the Content-Length header accordingly.

        :param value: The content size in bytes, or None to clear it
        """
        self._size = value
        if value is not None:
            self.headers["Content-Length"] = str(value)
        elif "Content-Length" in self.headers:
            del self.headers["Content-Length"]


__all__ = ["StreamResponse"]



# -*- coding: utf-8 -*-
# Copyright 2025, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from tempfile import TemporaryDirectory
from unittest.mock import patch

import yaml

from tests.context import (
    EODataAccessGateway,
    MisconfiguredError,
    PluginConfig,
    Provider,
    ProviderConfig,
    ProvidersDict,
    UnsupportedCollection,
    UnsupportedProvider,
    ValidationError,
)


class TestProvider(unittest.TestCase):
    """Test cases for the Provider class."""

    @classmethod
    def setUpClass(cls):
        """Set up test fixtures for the class."""
        super().setUpClass()
        cls.basic_config = {
            "name": "test_provider",
            "description": "Test provider for unit tests",
            "url": "https://test.example.com",
            "search": {"type": "StacSearch"},
            "products": {"S2_MSI_L1C": {"collection": "S2_MSI_L1C"}},
        }

    def setUp(self):
        super().setUp()
        # Mock home and eodag conf directory to tmp dir
        self.tmp_home_dir = TemporaryDirectory()
        self.expanduser_mock = patch(
            "os.path.expanduser", autospec=True, return_value=self.tmp_home_dir.name
        )
        self.expanduser_mock.start()

    def tearDown(self):
        super().tearDown()
        # stop Mock and remove tmp config dir
        self.expanduser_mock.stop()
        self.tmp_home_dir.cleanup()

    def test_provider_creation_from_dict(self):
        """Test Provider creation from dict config."""
        provider = Provider(self.basic_config)
        self.assertEqual(provider.name, "test_provider")
        self.assertIsInstance(provider.config, ProviderConfig)

    def test_provider_creation_from_provider_config(self):
        """Test Provider creation from ProviderConfig object."""
        config_obj = ProviderConfig.from_mapping(self.basic_config)
        provider = Provider(config_obj)
        self.assertEqual(provider.name, "test_provider")
        self.assertIsInstance(provider.config, ProviderConfig)

    def test_provider_creation_invalid_config(self):
        """Test Provider creation with invalid config."""
        with self.assertRaises(ValidationError):
            Provider("invalid_string")

        with self.assertRaises(ValidationError):
            Provider({"search": {"type": "StacSearch"}})  # missing name

    def test_provider_string_representations(self):
        """Test Provider string representations."""
        provider = Provider(self.basic_config)
        self.assertEqual(str(provider), "test_provider")
        self.assertEqual(repr(provider), "Provider('test_provider')")

    def test_provider_equality(self):
        """Test Provider equality."""
        provider1 = Provider(self.basic_config)
        provider2 = Provider(self.basic_config.copy())

        # Equal providers
        self.assertEqual(provider1, provider2)
        self.assertEqual(provider1, "test_provider")
        self.assertEqual(hash(provider1), hash(provider2))

        # Different providers
        config2 = self.basic_config.copy()
        config2["name"] = "different_provider"
        provider3 = Provider(config2)
        self.assertNotEqual(provider1, provider3)
        self.assertNotEqual(provider1, "different_provider")
        self.assertNotEqual(provider1, 123)

    def test_provider_basic_properties(self):
        """Test Provider basic properties."""
        provider = Provider(self.basic_config)
        self.assertEqual(provider.priority, 0)
        self.assertIsNone(provider.group)
        self.assertEqual(
            provider.collections_config, {"S2_MSI_L1C": {"collection": "S2_MSI_L1C"}}
        )
        self.assertEqual(provider.search_config.type, "StacSearch")
        self.assertIsNone(provider.api_config)
        self.assertIsNone(provider.download_config)

    def test_provider_custom_properties(self):
        """Test Provider properties with custom values."""
        config = self.basic_config.copy()
        config.update({"priority": 5, "group": "test_group"})
        provider = Provider(config)

        self.assertEqual(provider.priority, 5)
        self.assertEqual(provider.group, "test_group")

    def test_provider_fetchable(self):
        """Test Provider fetchable property."""
        provider = Provider(self.basic_config)
        self.assertTrue(provider.fetchable)

    def test_provider_delete_collection(self):
        """Test Provider delete_collection method."""
        provider = Provider(self.basic_config)

        # Successful deletion
        provider.delete_collection("S2_MSI_L1C")
        self.assertNotIn("S2_MSI_L1C", provider.collections_config)

        # Non-existent collection
        with self.assertRaises(UnsupportedCollection):
            provider.delete_collection("NONEXISTENT")

    def test_provider_sync_collections_strict(self):
        """Test Provider sync_collections in strict mode."""
        config = self.basic_config.copy()
        config["products"]["UNKNOWN_TYPE"] = {"collection": "UNKNOWN_TYPE"}
        provider = Provider(config)

        dag = EODataAccessGateway()
        with patch.object(provider, "delete_collection") as mock_delete:
            provider.sync_collections(dag, strict_mode=True)
            mock_delete.assert_called_once_with("UNKNOWN_TYPE")

    def test_provider_sync_collections_permissive(self):
        """Test Provider sync_collections in permissive mode."""
        config = self.basic_config.copy()
        config["products"]["UNKNOWN_TYPE"] = {"collection": "UNKNOWN_TYPE"}
        provider = Provider(config)

        dag = EODataAccessGateway()

        with patch.object(provider, "delete_collection") as mock_delete:
            provider.sync_collections(dag, strict_mode=False)
            mock_delete.assert_not_called()
            self.assertIn("UNKNOWN_TYPE", provider.collections_config)


class TestProviderConfig(unittest.TestCase):
    """Test cases for the ProviderConfig class."""

    def test_provider_config_creation(self):
        """Test ProviderConfig creation from mapping."""
        config_dict = {
            "name": "test_provider",
            "search": {"type": "StacSearch"},
        }
        config = ProviderConfig.from_mapping(config_dict)
        self.assertEqual(config.name, "test_provider")
        self.assertIsInstance(config.search, PluginConfig)

    def test_provider_config_validation_missing_name(self):
        """Test ProviderConfig validation with missing name."""
        with self.assertRaisesRegex(
            ValidationError, "Provider config must have name key"
        ):
            ProviderConfig.from_mapping({"search": {"type": "StacSearch"}})

    def test_provider_config_validation_missing_plugins(self):
        """Test ProviderConfig validation with missing plugins."""
        with self.assertRaisesRegex(
            ValidationError, "A provider must implement at least one plugin"
        ):
            ProviderConfig.from_mapping({"name": "test_provider"})

    def test_provider_config_validation_api_exclusivity(self):
        """Test that API plugin cannot coexist with other plugin types."""
        config = {
            "name": "test_provider",
            "api": {"type": "SomeApi"},
            "search": {"type": "StacSearch"},
        }
        with self.assertRaisesRegex(
            ValidationError, "Api plugin must not implement any other type"
        ):
            ProviderConfig.from_mapping(config)

    def test_provider_config_update(self):
        """Test ProviderConfig update operation."""
        config = ProviderConfig.from_mapping(
            {"name": "test_provider", "search": {"type": "StacSearch"}}
        )

        config.update({"description": "Updated description"})
        self.assertEqual(config.description, "Updated description")

    def test_provider_config_with_name(self):
        """Test ProviderConfig with_name method."""
        original_config = ProviderConfig.from_mapping(
            {
                "name": "original_provider",
                "description": "Original description",
                "priority": 5,
                "search": {"type": "StacSearch", "url": "https://example.com"},
                "products": {"S2_MSI_L1C": {"collection": "S2_MSI_L1C"}},
            }
        )

        # Create a copy with a new name
        new_config = original_config.with_name("new_provider")

        # Verify the new config has the updated name
        self.assertEqual(new_config.name, "new_provider")

        # Verify all other properties are preserved
        self.assertEqual(new_config.description, "Original description")
        self.assertEqual(new_config.priority, 5)
        self.assertIsInstance(new_config.search, PluginConfig)
        self.assertEqual(new_config.search.type, "StacSearch")
        self.assertEqual(new_config.search.url, "https://example.com")
        self.assertEqual(
            new_config.products, {"S2_MSI_L1C": {"collection": "S2_MSI_L1C"}}
        )

        # Verify the original config is unchanged
        self.assertEqual(original_config.name, "original_provider")

        # Verify they are separate objects (deep copy)
        self.assertIsNot(new_config, original_config)
        self.assertIsNot(new_config.search, original_config.search)

    def test_provider_config_from_yaml_requires_plugin_tag(self):
        """ProviderConfig.from_yaml must reject untagged plugin topics."""
        with self.assertRaisesRegex(
            MisconfiguredError, "plugin topic 'search' must be tagged with '!plugin'"
        ):
            yaml.load(
                """
!provider
name: test-provider
search:
  type: StacSearch
""",
                Loader=yaml.Loader,
            )


class TestProvidersDict(unittest.TestCase):
    """Test cases for the ProvidersDict class."""

    @classmethod
    def setUpClass(cls):
        """Set up test fixtures for the class."""
        super().setUpClass()
        cls.sample_configs = {
            "provider1": {
                "name": "provider1",
                "description": "First provider",
                "url": "https://provider1.example.com",
                "group": "group_a",
                "priority": 1,
                "search": {"type": "StacSearch"},
            },
            "provider2": {
                "name": "provider2",
                "description": "Second provider",
                "group": "group_b",
                "priority": 2,
                "search": {"type": "StacSearch"},
            },
        }
        cls.sample_providers = ProvidersDict.from_configs(cls.sample_configs)

    def test_providers_dict_creation(self):
        """Test ProvidersDict creation."""
        providers = ProvidersDict.from_configs(self.sample_configs)
        self.assertEqual(len(providers), 2)
        self.assertIn("provider1", providers)
        self.assertIn("provider2", providers)

    def test_providers_dict_contains(self):
        """Test ProvidersDict contains operation."""
        self.assertIn("provider1", self.sample_providers)
        self.assertNotIn("nonexistent", self.sample_providers)

        # Test with Provider object
        provider = self.sample_providers["provider1"]
        self.assertIn(provider, self.sample_providers)

    def test_providers_dict_attributes(self):
        """Test ProvidersDict provider attributes."""
        self.assertEqual(self.sample_providers["provider1"].title, "First provider")
        self.assertEqual(
            self.sample_providers["provider1"].url, "https://provider1.example.com"
        )
        self.assertEqual(self.sample_providers["provider2"].title, "Second provider")
        self.assertIsNone(self.sample_providers["provider2"].url)

    def test_providers_dict_basic_properties(self):
        """Test ProvidersDict basic properties."""
        self.assertEqual(len(self.sample_providers), 2)
        self.assertEqual(self.sample_providers.names, ["provider1", "provider2"])

        priorities = self.sample_providers.priorities
        self.assertEqual(priorities["provider1"], 1)
        self.assertEqual(priorities["provider2"], 2)

    def test_providers_dict_duplicate_provider(self):
        """Test ProvidersDict prevents duplicate providers."""
        new_provider = Provider({"name": "provider1", "search": {"type": "StacSearch"}})
        with self.assertRaisesRegex(ValueError, "Provider 'provider1' already exists"):
            self.sample_providers["provider1"] = new_provider

    def test_providers_dict_delete_provider(self):
        """Test ProvidersDict provider deletion."""
        # Create a fresh copy to avoid mutating shared state
        providers = ProvidersDict.from_configs(self.sample_configs)
        del providers["provider1"]
        self.assertNotIn("provider1", providers)
        self.assertEqual(len(providers), 1)

        # Test deleting non-existent provider
        with self.assertRaisesRegex(
            UnsupportedProvider, "Provider 'nonexistent' not found"
        ):
            del providers["nonexistent"]

    def test_providers_dict_filter_no_query(self):
        """Test ProvidersDict filter with no query."""
        results = list(self.sample_providers.filter())
        self.assertEqual(len(results), 2)

    def test_providers_dict_filter_with_query(self):
        """Test ProvidersDict filter with query."""
        results = self.sample_providers.filter("First")
        self.assertEqual(len(results), 1)
        self.assertEqual(results.names, ["provider1"])

    def test_providers_dict_filter_by_name_or_group(self):
        """Test ProvidersDict filter_by_name_or_group."""
        # Filter by name
        results = list(self.sample_providers.filter_by_name_or_group("provider1"))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].name, "provider1")

        # Filter by group
        results = list(self.sample_providers.filter_by_name_or_group("group_a"))
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].name, "provider1")

    def test_providers_dict_delete_collection(self):
        """Test ProvidersDict delete_collection method."""
        # Create a fresh copy to avoid mutating shared state
        providers = ProvidersDict.from_configs(self.sample_configs)
        # Add a collection first
        providers["provider1"].config.products = {"TEST": {"collection": "TEST"}}

        providers.delete_collection("provider1", "TEST")
        self.assertNotIn("TEST", providers["provider1"].collections_config)

        # Test with non-existent provider
        with self.assertRaises(UnsupportedProvider):
            providers.delete_collection("nonexistent", "TEST")

    def test_providers_dict_get_config(self):
        """Test ProvidersDict get_config method."""
        config = self.sample_providers.get_config("provider1")
        self.assertIsNotNone(config)
        self.assertEqual(config.name, "provider1")

        config = self.sample_providers.get_config("nonexistent")
        self.assertIsNone(config)

    @patch.dict("os.environ", {"EODAG_PROVIDERS_WHITELIST": "provider1"})
    def test_providers_dict_whitelist(self):
        """Test ProvidersDict whitelist filtering."""
        filtered = ProvidersDict._get_whitelisted_configs(self.sample_configs)
        self.assertEqual(set(filtered.keys()), {"provider1"})

    def test_providers_dict_invalid_config_handling(self):
        """Test ProvidersDict handles invalid configurations."""
        providers_dict = ProvidersDict()
        invalid_configs = {"invalid": {"description": "Missing required fields"}}

        with patch("eodag.api.provider.logger") as mock_logger:
            providers_dict.update_from_configs(invalid_configs)
            mock_logger.warning.assert_called()
            self.assertEqual(len(providers_dict), 0)



# -*- coding: utf-8 -*-
# Copyright 2026, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import datetime as dt
import logging
import math
from calendar import monthrange
from typing import Annotated, Any, Optional, cast

import requests
from pydantic.fields import FieldInfo
from pyproj import CRS, Transformer
from typing_extensions import get_args

from eodag.api.product._assets import AssetsDict
from eodag.api.product._product import EOProduct
from eodag.api.product.metadata_mapping import (
    mtd_cfg_as_conversion_and_querypath,
    properties_from_json,
)
from eodag.api.search_result import SearchResult
from eodag.plugins.search import PreparedSearch
from eodag.plugins.search.base import Search
from eodag.types import json_field_definition_to_python
from eodag.types.queryables import Queryables
from eodag.utils import DEFAULT_LIMIT, HTTP_REQ_TIMEOUT, USER_AGENT, deepcopy
from eodag.utils.cache import instance_cached_method
from eodag.utils.dates import parse_to_utc, to_iso_utc_string
from eodag.utils.exceptions import (
    MisconfiguredError,
    RequestError,
    TimeOutError,
    ValidationError,
)

logger = logging.getLogger("eodag.search.cop_ghsl")


def _convert_bbox_to_lonlat_mollweide(bbox: list[str]) -> list[float]:
    """
    convert a bbox from Mollweide coordinate system (metres)
    to WGS84 coordinate system (longitude and latitude)
    """
    bbox_int = [int(x.replace(" ", "")) for x in bbox]
    crs_mollweide = CRS("ESRI:54009")
    crs_wgs84 = CRS("WGS84")
    transformer = Transformer.from_crs(crs_from=crs_mollweide, crs_to=crs_wgs84)
    x1, y1 = bbox_int[:2]
    lat1, lon1 = transformer.transform(x1, y1)
    if math.isinf(lat1):
        # one corner is outside of the surface -> find latitude and take max longitude value
        lat1, _ = transformer.transform(0, y1)
        lon1 = -180 if x1 < 0 else 180
    x2, y2 = bbox_int[2:]
    lat2, lon2 = transformer.transform(x2, y2)
    if math.isinf(lat2):
        lat2, _ = transformer.transform(0, y2)
        lon2 = -180 if x1 < 0 else 180
    return [lon1, lat1, lon2, lat2]


def _convert_bbox_to_lonlat_EPSG3035(bbox: list[str]) -> list[float]:
    """
    convert a bbox from ETRS89/LAEA Europe (EPSG:3035) coordinate system (metres)
    to WGS84 coordinate system (longitude and latitude)
    """
    bbox_int = [int(x.replace(",", "")) for x in bbox]
    crs_3035 = CRS("3035")
    crs_wgs84 = CRS("WGS84")
    transformer = Transformer.from_crs(
        crs_from=crs_3035, crs_to=crs_wgs84, always_xy=True
    )
    x1, y1 = bbox_int[:2]
    lon1, lat1 = transformer.transform(x1, y1)
    x2, y2 = bbox_int[2:]
    lon2, lat2 = transformer.transform(x2, y2)
    return [lon1, lat1, lon2, lat2]


def _get_available_values_from_constraints(
    constraints: list[dict[str, Any]], filters: dict[str, Any], collection: str
) -> dict[str, list[Any]]:
    """get the available values for each parameter from the constraints"""
    available_values: dict[str, list[Any]] = {}
    constraint_keys = set([k for const in constraints for k in const.keys()])
    if (
        "tile_size" in constraint_keys
    ):  # tiles are available -> filtering by geometry is possible
        constraint_keys.add("geometry")
    not_found_keys = set(filters.keys()) - constraint_keys
    if "month" in not_found_keys and isinstance(filters["month"], list):
        # month added from datetime but filter not available
        filters.pop("month")
        not_found_keys.remove("month")
    if not_found_keys and not_found_keys != {"id"}:
        raise ValidationError(
            f"Parameters {not_found_keys} do not exist for collection {collection}; "
            f"available parameters: {constraint_keys}"
        )

    filtered_constraints = deepcopy(constraints)
    for filter_key, value in filters.items():
        available_values_key = []
        values_in_constraints = []
        for i, constraint in enumerate(constraints):
            if constraint not in filtered_constraints:
                continue
            if filter_key in constraint and isinstance(value, list):
                value_found = False
                for v in value:
                    if v in constraint[filter_key]:
                        value_found = True
                        values_in_constraints.append(v)
                if not value_found:
                    filtered_constraints.remove(constraint)
                    continue

            elif filter_key in constraint:
                available_values_key.extend(constraint[filter_key])
                if (
                    value not in constraint[filter_key]
                    and str(value) not in constraint[filter_key]
                ):
                    filtered_constraints.remove(constraint)
            if len(filtered_constraints) == 0:
                raise ValidationError(
                    f"Value {value} is not available for parameter {filter_key} with the given parameter values; "
                    f"available values: {set(available_values_key)}"
                )

        if not values_in_constraints and isinstance(value, list):
            constraints_values = []
            for const in constraints:
                constraints_values.extend(const[filter_key])
            raise ValidationError(
                f"No values for {filter_key} available in given range; available values "
                f"{sorted(set(constraints_values))}"
            )

    for constraint in filtered_constraints:
        for key, values in constraint.items():
            filter_values = values
            if key in filters and isinstance(filters[key], list):
                filter_values = list(set(filters[key]).intersection(set(values)))
            if key in available_values:
                available_values[key].extend(filter_values)
            else:
                available_values[key] = filter_values

    return available_values


def _replace_datetimes(params: dict[str, Any]):
    """replace datetimes by year/month"""
    start_date_str = params.pop("start_datetime", None)
    end_date_str = params.pop("end_datetime", None)
    if start_date_str and not end_date_str:
        end_date_str = start_date_str
    if end_date_str and not start_date_str:
        start_date_str = end_date_str

    if not start_date_str:
        return

    start_date = parse_to_utc(start_date_str)
    end_date = parse_to_utc(end_date_str)
    start_year = start_date.year
    end_year = end_date.year
    years = [str(y) for y in range(start_year, end_year + 1)]
    if "year" not in params:
        params["year"] = years

    if start_year == end_year and "month" not in params:
        # month is only used for collection where only one year is available
        start_month = start_date.month
        end_month = end_date.month
        months = [f"{m:02}" for m in range(start_month, end_month + 1)]
        params["month"] = months


class CopGhslSearch(Search):
    """
    Search plugin to fetch items from Copernicus Global Human Settlement Layer
    """

    def _check_input_parameters_valid(self, collection: str, params: Any):
        """
        Check if all required parameters are given and if the values are valid
        raises a ValidationError if this is not the case
        """
        constraints_data = self._fetch_constraints(collection)
        constraints_values = constraints_data["constraints"]
        # get available values - will raise error if wrong parameters or wrong parameter values in request
        grouped_by = params.pop("grouped_by", None)
        available_values = _get_available_values_from_constraints(
            constraints_values, params, collection
        )
        if grouped_by and grouped_by not in params:
            params[grouped_by] = available_values[grouped_by]
        missing_params = set(available_values.keys()) - set(params.keys())
        if missing_params:
            raise ValidationError(
                f"parameter(s) {missing_params} missing in the request"
            )
        # update lists in params with available values
        for param in params:
            if isinstance(params[param], list):
                params[param] = sorted(available_values[param])

    def _get_start_and_end_from_properties(
        self, properties: dict[str, Any]
    ) -> dict[str, str]:
        """get the start and end time from year/month in the properties or missionStart/EndDate"""
        if "month" in properties:
            start_date = dt.datetime(
                year=int(properties["year"]),
                month=int(properties["month"]),
                day=1,
                hour=0,
                minute=0,
                second=0,
            )
            end_day = monthrange(int(properties["year"]), int(properties["month"]))[1]
            end_date = dt.datetime(
                year=int(properties["year"]),
                month=int(properties["month"]),
                day=end_day,
                hour=23,
                minute=59,
                second=59,
            )
        elif "year" in properties:
            start_date = dt.datetime(
                year=int(properties["year"]), month=1, day=1, hour=0, minute=0, second=0
            )
            end_date = dt.datetime(
                year=int(properties["year"]),
                month=12,
                day=31,
                hour=23,
                minute=59,
                second=59,
            )
        else:
            interval = self.get_collection_cfg_value("extent")["temporal"]["interval"]
            start_date_str = interval[0][0]
            end_date_str = interval[0][1]
            return {"start_date": start_date_str, "end_date": end_date_str}

        result: dict[str, str] = {}
        # cast to tell the type checker that value won't be None here
        result["start_date"] = cast(str, to_iso_utc_string(start_date))
        result["end_date"] = cast(str, to_iso_utc_string(end_date))
        return result

    def _create_products_from_tiles(
        self,
        tiles: dict[str, list[dict[str, Any]]],
        unit: str,
        collection: str,
        params: dict[str, Any],
        additional_filter: Optional[str] = None,
        need_count=False,
    ) -> tuple[list[EOProduct], int]:
        """
        create EOProduct objects from the input parameters and the tiles containing bboxes
        if the bbox is given in metres, it is transformed to longitude and latitude
        """
        products = []
        metadata_mapping = params.pop("metadata_mapping", {})
        parsed_metadata_mapping = mtd_cfg_as_conversion_and_querypath(metadata_mapping)

        filter_geometry = params.pop("geometry", None)
        page = params.pop("page")
        per_page = params.pop("per_page")
        start_index = per_page * (page - 1)
        end_index = start_index + per_page - 1

        # parameters that need formatting
        dataset = metadata_mapping.get("dataset", None)
        if not dataset:
            raise MisconfiguredError(f"dataset mapping not available for {collection}")
        id_params = deepcopy(params)
        id_params["proj:code"] = id_params["proj:code"].replace("EPSG:", "")
        if "tile_size" in parsed_metadata_mapping:
            # format tile_size
            tile_size = properties_from_json(
                {"tile_size": params["tile_size"]}, parsed_metadata_mapping
            )["tile_size"]
            params.update({"tile_size": tile_size})

        # additional filter
        if additional_filter:
            add_filter_value = params.pop(additional_filter)
            if add_filter_value == "TOTAL":
                params.update({"add_filter": ""})
            else:
                params.update({"add_filter": add_filter_value})

        if isinstance(params["year"], int) or isinstance(params["year"], str):
            list_years = [str(params["year"])]
        else:
            list_years = params["year"]
        current_index = 0
        for year in list_years:
            properties = deepcopy(params)
            properties["order:status"] = "succeeded"
            properties["start_datetime"] = to_iso_utc_string(
                dt.datetime(year=int(year), month=1, day=1)
            )
            properties["end_datetime"] = to_iso_utc_string(
                dt.datetime(
                    year=int(year), month=12, day=31, hour=23, minute=59, second=59
                )
            )
            properties["year"] = year

            # information for id and download path
            id_params["year"] = year
            id_params = {k: str(v) for k, v in id_params.items()}
            product_id_base = (
                collection + "__" + "_".join(v for v in id_params.values() if v)
            )

            params["year"] = year
            params = {k.replace(":", "_"): v for k, v in params.items()}
            # remove EPSG prefix for download url
            params["proj_code"] = params["proj_code"].replace("EPSG:", "")
            dataset = dataset.format(**params)
            dataset = dataset.replace(
                "__", "_"
            )  # in case additional filter value is empty

            # create items from tiles
            for tile in tiles[year]:
                if not tile:  # empty grid position
                    continue
                # get geometry from tile
                if unit == "lat/lon":  # bbox is given as latitude/longitude
                    properties["geometry"] = tile["BBox"]
                elif unit == "metres" and "BBox" in tile:  # bbox is given in metres
                    bbox_lon_lat = _convert_bbox_to_lonlat_mollweide(tile["BBox"])
                    properties["geometry"] = bbox_lon_lat
                else:  # ETRS89/LAEA Europe coordinate system
                    bbox_lon_lat = _convert_bbox_to_lonlat_EPSG3035(tile["BBox_3035"])
                    properties["geometry"] = bbox_lon_lat
                # create id
                product_id = f"{product_id_base}__{tile['tileID']}"
                properties["id"] = properties["title"] = product_id
                download_link = metadata_mapping.get("eodag:download_link").format(
                    dataset=dataset, tile_id=tile["tileID"]
                )
                properties["eodag:download_link"] = download_link
                product = EOProduct(
                    provider="cop_ghsl", properties=properties, collection=collection
                )
                if not filter_geometry or filter_geometry.intersects(product.geometry):
                    if current_index >= start_index and current_index <= end_index:
                        products.append(product)
                    elif current_index > end_index and not need_count:
                        break
                    current_index += 1

        return products, current_index

    def _create_products_without_tiles(
        self, collection: str, prep: PreparedSearch, filter_params: dict[str, Any]
    ) -> tuple[list[EOProduct], Optional[int]]:
        filters = deepcopy(filter_params)
        default_geometry = getattr(self.config, "metadata_mapping")[
            "eodag:default_geometry"
        ]
        properties = {}
        properties["geometry"] = default_geometry[1]
        properties["order:status"] = "succeeded"
        if "proj:code" in filters:
            filters["proj:code"] = filters["proj:code"].replace("EPSG:", "")
        collection_config = self.config.products.get(collection, {})
        download_link = collection_config.get("metadata_mapping", {}).get(
            "eodag:download_link", None
        )
        if not download_link:
            raise MisconfiguredError(
                f"Download link configuration missing for collection {collection}"
            )

        # collection with assets mapping
        assets_mapping = filters.pop("assets_mapping", None)
        products = []
        per_page = getattr(prep, "limit", DEFAULT_LIMIT)
        page = getattr(prep, "PAGE", 1)
        start_index = per_page * (page - 1)
        end_index = start_index + per_page - 1
        grouped_by = filters.pop("grouped_by", None)
        if grouped_by:  # dataset with several files differentiated by one parameter
            format_params = {k: str(v) for k, v in filters.items() if v}
            format_params.pop("metadata_mapping", None)
            grouped_by_values = filters[grouped_by]
            if isinstance(grouped_by_values, str) or isinstance(grouped_by_values, int):
                grouped_by_values = [grouped_by_values]
            num_products = len(grouped_by_values)
            for i, value in enumerate(grouped_by_values):
                if i < start_index:
                    continue
                filters[grouped_by] = format_params[grouped_by] = str(value)
                product_id = collection + "__" + "_".join(format_params.values())
                properties["id"] = properties["title"] = product_id
                properties.update(format_params)
                if "proj:code" in filter_params:
                    properties["proj:code"] = filter_params["proj:code"]
                properties["eodag:download_link"] = download_link.format(
                    **format_params
                )
                datetimes = self._get_start_and_end_from_properties(format_params)
                properties["start_datetime"] = datetimes["start_date"]
                properties["end_datetime"] = datetimes["end_date"]
                properties[grouped_by] = value
                product = EOProduct(
                    provider="cop_ghsl", properties=properties, collection=collection
                )
                if assets_mapping:  # item with several assets
                    assets = AssetsDict(product=product)
                    for key, mapping in assets_mapping.items():
                        filters = {k.replace(":", "_"): v for k, v in filters.items()}
                        download_link = mapping["href"].format(**filters)
                        assets.update(
                            {
                                key: {
                                    "href": download_link,
                                    "title": mapping["title"],
                                    "type": mapping["type"],
                                }
                            }
                        )
                    product.assets = assets
                products.append(product)
                if i == end_index:
                    break
        else:  # collection with only one file to download
            product_id = f"{collection}_ALL"
            properties["id"] = properties["title"] = product_id
            datetimes = self._get_start_and_end_from_properties(properties)
            properties["start_datetime"] = datetimes["start_date"]
            properties["end_datetime"] = datetimes["end_date"]
            properties["eodag:download_link"] = download_link
            product = EOProduct(
                provider="cop_ghsl", properties=properties, collection=collection
            )
            products.append(product)
            num_products = 1
        if prep.count:
            return products, num_products
        else:
            return products, None

    def _get_tile_from_product_id(
        self, query_params: dict[str, Any]
    ) -> Optional[tuple[dict[str, list[dict[str, Any]]], str]]:
        """fetch the tile for a specific product id from the provider
        returns a a dict with a list of length 1 to simplify further processing
        """
        product_id = query_params.pop("id")
        collection = query_params["collection"]
        tile_id = product_id.split("__")[-1]
        filter_part = product_id.split("__")[1]
        constraints_values = self._fetch_constraints(collection)["constraints"]
        available_values = _get_available_values_from_constraints(
            constraints_values, {}, collection
        )
        collection_config = deepcopy(self.config.products.get(collection, {}))
        for key, values in available_values.items():
            for value in values:
                param_value = value
                if key == "proj:code":
                    param_value = value.replace("EPSG:", "")
                if param_value in filter_part:
                    query_params[key] = value
                    break
        tiles_or_none = self._get_tiles_for_filters(collection_config, query_params)
        if tiles_or_none:
            tiles, unit = tiles_or_none
            matching_tile = [
                tile
                for tile in tiles[query_params["year"]]
                if tile and tile["tileID"] == tile_id
            ]
            return {query_params["year"]: matching_tile}, unit
        else:
            return None

    def _get_tiles_for_filters(
        self, collection_config: dict[str, Any], params: dict[str, Any]
    ) -> Optional[tuple[dict[str, list[dict[str, Any]]], str]]:
        """fetch the tiles matching the given filters from the provider"""

        logger.debug(f"get tiles for filter parameters {params}")
        collection = params.pop("collection")
        ssl_verify = getattr(self.config, "ssl_verify", True)
        timeout = getattr(self.config, "timeout", HTTP_REQ_TIMEOUT)

        # update filters with values from collection mapping
        provider_product_type = collection_config.pop("_collection", None)
        if not provider_product_type:
            raise MisconfiguredError(
                f"provider collection mapping not available for {collection}"
            )

        params.pop("geometry", None)
        params.update(collection_config)
        params.pop("metadata_mapping", None)
        params.pop("assets_mapping", None)

        self._check_input_parameters_valid(collection, params)
        # update parameters based on changes during validation

        # fetch available tiles based on filters
        if "year" not in params:
            logger.warning(f"no tiles available for {collection}")
            return None
        if isinstance(params["year"], int) or isinstance(params["year"], str):
            list_years = [str(params["year"])]
        else:
            list_years = params["year"]
        all_tiles = {}
        # remove EPSG prefix for request to provider
        proj_code = params["proj:code"].replace("EPSG:", "")
        for year in list_years:
            try:
                filter_str = (
                    f"{provider_product_type}_{year}_"
                    f"{params['tile_size']}_{proj_code}"
                )
            except KeyError:
                logger.warning(f"no tiles available for {collection}")
                return None
            tiles_url = self.config.api_endpoint + "/tilesDLD_" + filter_str + ".json"
            try:
                res = requests.get(
                    tiles_url, verify=ssl_verify, timeout=timeout, headers=USER_AGENT
                )
                if res.status_code == 404:
                    return None
                res.raise_for_status()
                tiles = res.json()["grid"]
                if params["proj:code"] == "EPSG:3035":
                    tiles = []
                    for t_id, bbox in res.json()["BBoxes"].items():
                        tiles.append({"tileID": t_id, "BBox_3035": bbox})
                all_tiles[year] = tiles
                unit = res.json().get("unit", "")
            except requests.exceptions.Timeout as exc:
                raise TimeOutError(exc, timeout=timeout) from exc
            except requests.exceptions.RequestException as exc:
                raise RequestError.from_error(
                    exc, f"Unable to fetch {tiles_url}"
                ) from exc

        return all_tiles, unit

    def query(
        self,
        prep: PreparedSearch = PreparedSearch(),
        **kwargs: Any,
    ) -> SearchResult:
        """
        Implementation of search for the Copernicus GHSL provider

        :param prep: object containing search parameters
        :param kwargs: additional search arguments
        :returns: list of products and total number of products
        """
        page = int(getattr(prep, "next_page_token") or "1")
        limit = getattr(prep, "limit") or DEFAULT_LIMIT
        number_matched = kwargs.pop("number_matched", None)

        # get year/month from start/end time if not given separately
        _replace_datetimes(kwargs)

        # search params for SearchResult
        search_params = deepcopy(kwargs)
        search_params["limit"] = limit

        collection = kwargs.get("collection", None)
        if not collection:
            collection = kwargs["collection"] = prep.collection
        if not isinstance(collection, str):
            raise MisconfiguredError("invalid collection %s", collection)
        collection_config = deepcopy(self.config.products.get(collection, {}))
        if "id" in kwargs and "ALL" not in kwargs["id"]:
            tiles_or_none = self._get_tile_from_product_id(kwargs)
        else:
            tiles_or_none = self._get_tiles_for_filters(collection_config, kwargs)
        if tiles_or_none:
            tiles, unit = tiles_or_none
        else:
            kwargs.update(collection_config)
            products, count = self._create_products_without_tiles(
                collection, prep, kwargs
            )
            return SearchResult(
                products=products,
                search_params=search_params,
                number_matched=count,
                next_page_token_key="page",
                next_page_token=str(page + 1),
                raise_errors=True,
            )

        # create products from tiles
        kwargs.update(collection_config)
        kwargs["page"] = page
        kwargs["per_page"] = limit
        constraints_filters = self._fetch_constraints(collection)
        additional_filter = constraints_filters.get("additional_filter")
        need_count = prep.count and not number_matched
        products, count = self._create_products_from_tiles(
            tiles,
            unit,
            collection,
            kwargs,
            additional_filter=additional_filter,
            need_count=need_count,
        )

        if prep.count:
            if number_matched:
                total_items = number_matched
            else:
                total_items = count
        else:
            total_items = None
        return SearchResult(
            products=products,
            search_params=search_params,
            number_matched=total_items,
            next_page_token_key="page",
            next_page_token=str(page + 1),
            raise_errors=True,
        )

    @instance_cached_method()
    def _fetch_constraints(self, collection: str) -> dict[str, Any]:
        logger.debug(f"fetching constraints for {collection}")
        constraints_url = self.config.discover_queryables["constraints_url"].format(
            collection=collection
        )
        timeout = getattr(self.config, "timeout", HTTP_REQ_TIMEOUT)
        try:
            res = requests.get(constraints_url, timeout=timeout, headers=USER_AGENT)
            if res.status_code == 404:
                logger.warning(f"no constraints found for {collection}")
                return {"constraints": {}}
            res.raise_for_status()
            return res.json()
        except requests.exceptions.Timeout as exc:
            raise TimeOutError(exc, timeout=timeout) from exc
        except requests.exceptions.RequestException as exc:
            raise RequestError.from_error(
                exc, f"Unable to fetch constraints from {constraints_url}"
            ) from exc

    def discover_queryables(
        self, **kwargs
    ) -> Optional[dict[str, Annotated[Any, FieldInfo]]]:
        """Create queryables list based on constraints

        :param kwargs: additional filters for queryables (`collection` and other search
                       arguments)
        :returns: queryable parameters dict
        """

        collection = kwargs.pop("collection")
        kwargs.pop("_collection")
        grouped_by = kwargs.pop("grouped_by", None)
        _replace_datetimes(kwargs)
        constraints_values = self._fetch_constraints(collection)["constraints"]
        available_values = _get_available_values_from_constraints(
            constraints_values, kwargs, collection
        )
        queryables = {}
        for name, values in available_values.items():
            required = True
            if name == grouped_by:
                required = False
            queryables[name] = Annotated[
                get_args(
                    json_field_definition_to_python(
                        {"type": "string", "title": name, "enum": values},
                        default_value=kwargs.get(name, None),
                        required=required,
                    )
                )
            ]
        # add datetimes queryables if year filter is available
        if "year" in available_values:
            queryables.update(
                {
                    "start": Queryables.get_with_default(
                        "start", kwargs.get("start_datetime")
                    ),
                    "end": Queryables.get_with_default(
                        "end",
                        kwargs.get("end_datetime"),
                    ),
                }
            )
        # add geometry queryable if there are tiles
        if "tile_size" in available_values and not grouped_by:
            queryables.update(
                {
                    "geom": Queryables.get_with_default("geom", None),
                }
            )

        return queryables



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""EODAG product package"""

from __future__ import annotations

from typing import TYPE_CHECKING

from eodag.utils.deserialize import unregistered_product_from_item

if TYPE_CHECKING:
    from eodag.api.product._assets import Asset, AssetsDict
    from eodag.api.product._product import EOProduct

# exportable content
__all__ = ["Asset", "AssetsDict", "EOProduct", "unregistered_product_from_item"]

# Lazy imports (PEP 562) — defer heavy _product / _assets loading
_LAZY_IMPORTS: dict[str, tuple[str, str]] = {
    "EOProduct": ("._product", "EOProduct"),
    "Asset": ("._assets", "Asset"),
    "AssetsDict": ("._assets", "AssetsDict"),
}


def _resolve(name: str):
    """Resolve a lazy import, trying eodag-cube first for EOProduct/Asset/AssetsDict."""
    module_path, attr_name = _LAZY_IMPORTS[name]
    from importlib import import_module

    # Try eodag_cube first (provides extended classes with xarray/rasterio support).
    # Users just need ``pip install eodag_cube`` — no other change required.
    try:
        cube_module = import_module("eodag_cube.api.product")
        value = getattr(cube_module, attr_name, None)
        if value is not None:
            globals()[name] = value
            return value
    except ImportError:
        pass

    # Fallback to eodag's own modules
    module = import_module(module_path, __name__)
    value = getattr(module, attr_name)
    globals()[name] = value
    return value


def __getattr__(name: str):
    if name in _LAZY_IMPORTS:
        return _resolve(name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return __all__



# -*- coding: utf-8 -*-
# Copyright 2026, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from urllib.parse import parse_qs, urljoin

import requests
from lxml import html
from requests.auth import AuthBase

from eodag.plugins.authentication.base import Authentication
from eodag.utils.exceptions import AuthenticationError, MisconfiguredError


class EOIAMAuth(Authentication):
    """
    Authentication plugin for EOIAM.
    """

    def __init__(self, provider, config) -> None:
        """Initialize the plugin with provider and config,
        and set up a requests.Session for SAML login."""
        super().__init__(provider, config)
        self.session = requests.Session()

    def validate_config_credentials(self) -> None:
        """Validate configured credentials"""
        required = ["username", "password"]
        missing = [
            k for k in required if k not in getattr(self.config, "credentials", {})
        ]
        if missing:
            msg = f"Missing credentials for {self.provider}: {', '.join(missing)}"
            raise MisconfiguredError(msg)

    def authenticate(self) -> AuthBase:
        """Return a requests.AuthBase object using the session with SAML login."""
        self.validate_config_credentials()
        return _EOIAMSessionAuth(self)

    def _extract_input_value(self, tree, name: str) -> str:
        """Extract the value of an input field from the HTML tree."""
        inputs = tree.xpath(f"//input[@name='{name}']")
        if not inputs:
            msg = f"{name} input not found"
            raise MisconfiguredError(msg)
        value = inputs[0].get("value")
        if not value:
            msg = f"{name} has no value"
            raise MisconfiguredError(msg)
        return value

    def _extract_first_form(self, tree: html.HtmlElement) -> html.HtmlElement:
        """Extract the first form from the HTML tree."""
        forms = tree.xpath("//form")
        if not forms:
            raise MisconfiguredError("Form not found")
        return forms[0]

    def _resolve_action(self, form: html.HtmlElement, base_url: str) -> str:
        """Resolve the action URL of a form relative to the base URL."""
        action = form.get("action")
        if not action:
            raise MisconfiguredError("Form action not found")
        return urljoin(base_url, action)

    def _login_from_html(self, html_content: str, req_url: str) -> requests.Response:
        """Perform SAML login from HTML page."""
        creds = self.config.credentials
        username = creds["username"]
        password = creds["password"]
        base_url = self.config.auth_uri

        # Parse the HTML to extract the session key and form action
        tree = html.fromstring(html_content)

        session_key = self._extract_input_value(tree, "sessionDataKey")

        form = self._extract_first_form(tree)
        idp_url = self._resolve_action(form, base_url)

        # Submit credentials
        payload = {
            "tocommonauth": "true",
            "username": username,
            "password": password,
            "sessionDataKey": session_key,
        }
        resp = self.session.post(idp_url, data=payload, allow_redirects=True)
        resp.raise_for_status()

        if "consent.do" in resp.url:
            redirect_url = resp.url
            service_names = parse_qs(redirect_url).get("sp", [""])
            service_name = service_names[0] if service_names else ""
            msg = (
                f"Consent required for service {service_name}, "
                f"please fill the following form and try again {req_url}"
            )
            raise AuthenticationError(msg)

        if "login.do" in resp.url:
            raise MisconfiguredError("Login failed: please check your credentials")

        if "Earth Observation Identity and Access Management System" in resp.text:
            msg = f"Login failed: {resp.url}"
            raise MisconfiguredError(msg)

        # Extract SAML form
        tree = html.fromstring(resp.text)
        form = self._extract_first_form(tree)

        saml_url = self._resolve_action(form, base_url)
        saml_data = {
            inp.get("name"): inp.get("value")
            for inp in form.xpath(".//input[@name]")
            if inp.get("name") and inp.get("value")
        }

        # Submit SAML response
        resp_post = self.session.post(saml_url, data=saml_data, allow_redirects=False)

        if not resp_post.is_redirect:
            msg = f"Unexpected response after SAML login: {resp_post.status_code}"
            raise AuthenticationError(msg)

        final_url = resp_post.headers.get("Location")
        if not final_url:
            raise AuthenticationError("Final redirect URL not found after SAML login")

        resp_final = self.session.get(final_url, stream=True)

        content_type = resp_final.headers.get("Content-Type", "")

        if content_type.startswith(("text/html", "text/xml")):
            error_text = resp_final.text

            if "wants to access your account" in error_text:
                msg = (
                    "Consent required: please log in to the EOIAM portal "
                    f"and grant consent through this link {final_url}"
                )
                raise AuthenticationError(msg)

            if (
                "not yet performed the necessary steps in order to access this data."
                in error_text
            ):
                msg = (
                    f"Data access request required: please log in to the EOIAM portal "
                    f"and request access to the data through this link {final_url}"
                )
                raise AuthenticationError(msg)

            raise AuthenticationError("Unexpected HTML response after SAML login")

        return resp_final


class _EOIAMSessionAuth(AuthBase):
    """AuthBase wrapper using a requests.Session with lazy SAML login."""

    def __init__(self, auth_plugin: EOIAMAuth):
        """Initialize with the EOIAMAuth plugin to access its session and login method."""
        self.auth_plugin = auth_plugin

    def __call__(self, request):
        """
        This is called by requests before sending a request.
        We use the session's get/post to ensure login happens if needed.
        """
        session = self.auth_plugin.session
        try:
            resp = session.get(request.url, allow_redirects=True)
            if "Earth Observation Identity and Access Management System" in resp.text:
                resp = self.auth_plugin._login_from_html(resp.text, request.url)

            # Copy cookies from session to the request
            request.prepare_cookies(self.auth_plugin.session.cookies)

            return request
        finally:
            self.auth_plugin.session = requests.Session()



# -*- coding: utf-8 -*-
# Copyright 2025, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import os
import tempfile
import traceback
from collections import UserDict
from inspect import isclass
from textwrap import shorten
from typing import (
    TYPE_CHECKING,
    Any,
    Iterator,
    Mapping,
    Optional,
    Union,
    get_type_hints,
)

import yaml

from eodag.api.collection import Collection
from eodag.api.product.metadata_mapping import (
    NOT_AVAILABLE,
    mtd_cfg_as_conversion_and_querypath,
)
from eodag.config import PluginConfig, credentials_in_auth, load_stac_provider_config
from eodag.utils import (
    GENERIC_COLLECTION,
    STAC_SEARCH_PLUGINS,
    cast_scalar_value,
    deepcopy,
    merge_mappings,
    slugify,
    update_nested_dict,
)
from eodag.utils.exceptions import (
    MisconfiguredError,
    UnsupportedCollection,
    UnsupportedProvider,
    ValidationError,
)
from eodag.utils.free_text_search import compile_free_text_query
from eodag.utils.repr import dict_to_html_table, str_as_href

if TYPE_CHECKING:
    from typing_extensions import Self

    from eodag.api.core import EODataAccessGateway

logger = logging.getLogger("eodag.provider")

AUTH_TOPIC_KEYS = ("auth", "search_auth", "download_auth")
PLUGINS_TOPICS_KEYS = ("api", "search", "download") + AUTH_TOPIC_KEYS


class ProviderConfig(yaml.YAMLObject):
    """EODAG configuration for a provider.

    :param name: The name of the provider
    :param priority: (optional) The priority of the provider while searching a product.
                     Lower value means lower priority. (Default: 0)
    :param roles: The roles of the provider (e.g. "host", "producer", "licensor", "processor")
    :param description: (optional) A short description of the provider
    :param url: URL to the webpage representing the provider
    :param api: (optional) The configuration of a plugin of type Api
    :param search: (optional) The configuration of a plugin of type Search
    :param products: (optional) The collections supported by the provider
    :param download: (optional) The configuration of a plugin of type Download
    :param auth: (optional) The configuration of a plugin of type Authentication
    :param search_auth: (optional) The configuration of a plugin of type Authentication for search
    :param download_auth: (optional) The configuration of a plugin of type Authentication for download
    :param kwargs: Additional configuration variables for this provider
    """

    name: str
    group: str
    priority: int = 0
    roles: list[str]
    description: str
    url: str
    api: PluginConfig
    search: PluginConfig
    products: dict[str, Any]
    download: PluginConfig
    auth: PluginConfig
    search_auth: PluginConfig
    download_auth: PluginConfig

    yaml_loader = yaml.Loader
    yaml_dumper = yaml.SafeDumper
    yaml_tag = "!provider"

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Apply defaults when building from yaml."""
        self.__dict__.update(state)
        self._apply_defaults()

    def __contains__(self, key):
        """Check if a key is in the ProviderConfig."""
        return key in self.__dict__

    @classmethod
    def from_yaml(cls, loader: yaml.Loader, node: Any) -> Iterator[Self]:
        """Build a :class:`~eodag.api.provider.ProviderConfig` from Yaml"""
        cls.validate(tuple(node_key.value for node_key, _ in node.value))
        for node_key, node_value in node.value:
            if node_key.value == "name":
                node_value.value = slugify(node_value.value).replace("-", "_")
            elif node_key.value in PLUGINS_TOPICS_KEYS:
                if node_value.tag != PluginConfig.yaml_tag:
                    msg = "Provider plugin topic '%s' must be tagged with '%s'" % (
                        node_key.value,
                        PluginConfig.yaml_tag,
                    )
                    raise MisconfiguredError(msg)
        return loader.construct_yaml_object(node, cls)

    @classmethod
    def from_mapping(cls, mapping: dict[str, Any]) -> Self:
        """Build a :class:`~eodag.api.provider.ProviderConfig` from a mapping"""
        cls.validate(mapping)
        # Create a deep copy to avoid modifying the input dict or its nested structures
        mapping_copy = deepcopy(mapping)
        for key in PLUGINS_TOPICS_KEYS:
            if not (_mapping := mapping_copy.get(key)):
                continue

            if not isinstance(_mapping, dict):
                _mapping = _mapping.__dict__

            mapping_copy[key] = PluginConfig.from_mapping(_mapping)
        c = cls()
        c.__dict__.update(mapping_copy)
        c._apply_defaults()
        return c

    @staticmethod
    def validate(config_keys: Union[tuple[str, ...], dict[str, Any]]) -> None:
        """Validate a :class:`~eodag.api.provider.ProviderConfig`

        :param config_keys: The configurations keys to validate
        """
        if "name" not in config_keys:
            raise ValidationError("Provider config must have name key")
        if not any(k in config_keys for k in PLUGINS_TOPICS_KEYS):
            raise ValidationError("A provider must implement at least one plugin")
        non_api_keys = [k for k in PLUGINS_TOPICS_KEYS if k != "api"]
        if "api" in config_keys and any(k in config_keys for k in non_api_keys):
            raise ValidationError(
                "A provider implementing an Api plugin must not implement any other "
                "type of plugin"
            )

    def update(self, config: Union[Self, dict[str, Any]]) -> None:
        """Update the configuration parameters with values from `mapping`

        :param config: The config from which to override configuration parameters
        """
        source = config if isinstance(config, dict) else config.__dict__

        merge_mappings(
            self.__dict__,
            {
                key: value
                for key, value in source.items()
                if key not in PLUGINS_TOPICS_KEYS and value is not None
            },
        )
        for key in PLUGINS_TOPICS_KEYS:
            current_value: Optional[PluginConfig] = getattr(self, key, None)
            config_value = source.get(key, {})
            if current_value is not None:
                current_value |= config_value
            elif isinstance(config_value, PluginConfig):
                setattr(self, key, config_value)
            elif config_value:
                try:
                    setattr(self, key, PluginConfig.from_mapping(config_value))
                except ValidationError as e:
                    logger.warning(
                        (
                            "Could not add %s Plugin config to %s configuration: %s. "
                            "Try updating existing %s Plugin configs instead."
                        ),
                        key,
                        self.name,
                        str(e),
                        ", ".join([k for k in PLUGINS_TOPICS_KEYS if hasattr(self, k)]),
                    )
        self._apply_defaults()

    def with_name(self, new_name: str) -> Self:
        """Create a copy of this :class:`~eodag.api.provider.ProviderConfig` with a different name.

        :param new_name: The new name for the provider config.
        :return: A new ProviderConfig instance with the updated name.
        """
        config_dict = self.__dict__.copy()
        config_dict["name"] = new_name

        for key in PLUGINS_TOPICS_KEYS:
            if key in config_dict and isinstance(config_dict[key], PluginConfig):
                config_dict[key] = config_dict[key].__dict__

        return self.__class__.from_mapping(config_dict)

    def _apply_defaults(self: Self) -> None:
        """Applies some default values to provider config."""
        stac_search_default_conf = load_stac_provider_config()

        # For the provider, set the default output_dir of its download plugin
        # as tempdir in a portable way
        for download_topic_key in ("download", "api"):
            if download_topic_key in vars(self):
                download_conf = getattr(self, download_topic_key)
                if not getattr(download_conf, "output_dir", None):
                    download_conf.output_dir = tempfile.gettempdir()
                if not getattr(download_conf, "delete_archive", None):
                    download_conf.delete_archive = True

        try:
            if (
                stac_search_default_conf is not None
                and self.search
                and self.search.type in STAC_SEARCH_PLUGINS
            ):
                # search config set to stac defaults overriden with provider config
                per_provider_stac_provider_config = deepcopy(stac_search_default_conf)
                self.search.__dict__ = update_nested_dict(
                    per_provider_stac_provider_config["search"],
                    self.search.__dict__,
                    allow_empty_values=True,
                )
        except AttributeError:
            pass


class Provider:
    """
    Represents a data provider with its configuration and utility methods.

    :param config: Provider configuration as :meth:`~eodag.api.provider.ProviderConfig` instance or :class:`dict`
    :param collections_fetched: Flag indicating whether collections have been fetched

    Example
    -------

    >>> from eodag.api.provider import Provider
    >>> config = {
    ...     'name': 'example_provider',
    ...     'description': 'Example provider for testing',
    ...     'search': {'type': 'StacSearch'},
    ...     'products': {'S2_MSI_L1C': {'_collection': 'S2_MSI_L1C'}}
    ... }
    >>> provider = Provider(config)
    >>> provider.name
    'example_provider'
    >>> 'S2_MSI_L1C' in provider.collections_config
    True
    >>> provider.priority  # Default priority
    0
    """

    _name: str
    _config: ProviderConfig
    collections_fetched: bool

    def __init__(self, config: Union[ProviderConfig, dict[str, Any]]):
        """Initialize provider with configuration."""
        if isinstance(config, dict):
            self._config = ProviderConfig.from_mapping(config)
        elif isinstance(config, ProviderConfig):
            self._config = config
        else:
            msg = f"Unsupported config type: {type(config)}. Expected ProviderConfig or dict."
            raise ValidationError(msg)

        self._name = self._config.name
        self.collections_fetched = False

    def __str__(self) -> str:
        """Return the provider's name as string."""
        return self.name

    def __repr__(self) -> str:
        """Return a string representation of the Provider."""
        return f"Provider('{self.name}')"

    def __eq__(self, other: object):
        """Compare providers by name or with a string."""
        if isinstance(other, Provider):
            return self.name == other.name

        elif isinstance(other, str):
            return self.name == other

        return False

    def __hash__(self):
        """Hash based on provider name, for use in sets/dicts."""
        return hash(self.name)

    def _repr_html_(self, embedded: bool = False) -> str:
        """HTML representation for Jupyter/IPython display."""
        group_display = f" ({self.group})" if self.group else ""
        thead = (
            f"""<thead><tr><td style='text-align: left; color: grey;'>
                {type(self).__name__}("<span style='color: black'>{self.name}{group_display}</span>")</td></tr></thead>
            """
            if not embedded
            else ""
        )
        tr_style = "style='background-color: transparent;'" if embedded else ""

        summaries = {
            "name": self.name,
            "title": self.config.description or "",
            "url": self.config.url or "",
            "priority": self.priority,
        }
        if self.group:
            summaries["group"] = self.group

        col_html_table = dict_to_html_table(summaries, depth=1, brackets=False)

        return (
            f"<table>{thead}<tbody>"
            f"<tr {tr_style}><td style='text-align: left;'>"
            f"{col_html_table}</td></tr>"
            "</tbody></table>"
        )

    @property
    def config(self) -> ProviderConfig:
        """
        Provider configuration (read-only assignment).

        To update configuration safely, use :meth:`~eodag.api.provider.Provider.update_from_config`
        which handles metadata mapping and other provider-specific logic.

        Note: Direct config modification (``config.update()``, ``config.name = ...``)
        bypasses important provider validation.
        """
        return self._config

    @property
    def name(self) -> str:
        """The name of the provider."""
        return self._name

    @property
    def title(self) -> Optional[str]:
        """The title of the provider."""
        return getattr(self.config, "description", None)

    @property
    def url(self) -> Optional[str]:
        """The url of the provider."""
        return getattr(self.config, "url", None)

    @property
    def collections_config(self) -> dict[str, Any]:
        """Return the collections configuration dictionary for this provider."""
        return getattr(self.config, "products", {})

    @property
    def priority(self) -> int:
        """Return the provider's priority (default: 0)."""
        return self.config.priority

    @property
    def group(self) -> Optional[str]:
        """Return the provider's group, if any."""
        return getattr(self.config, "group", None)

    @property
    def search_config(self) -> Optional[PluginConfig]:
        """Return the search plugin config, if any."""
        return getattr(self.config, "search", None) or getattr(self.config, "api", None)

    @property
    def fetchable(self) -> bool:
        """Return True if the provider can fetch collections."""
        return bool(
            getattr(self.search_config, "discover_collections", {}).get("fetch_url")
        )

    @property
    def unparsable_properties(self) -> set[str]:
        """Return set of unparsable properties from
        :attr:`~eodag.config.PluginConfig.DiscoverCollections.generic_collection_unparsable_properties`, if any.
        """
        if not self.fetchable or self.search_config is None:
            return set()

        props = getattr(
            getattr(self.search_config, "discover_collections", None),
            "generic_collection_unparsable_properties",
            {},
        )
        return set(props.keys()) if isinstance(props, dict) else set()

    @property
    def api_config(self) -> Optional[PluginConfig]:
        """Return the api plugin config, if any."""
        return getattr(self.config, "api", None)

    @property
    def download_config(self) -> Optional[PluginConfig]:
        """Return the download plugin config, if any."""
        return getattr(self.config, "download", None)

    def _get_auth_confs_with_credentials(self) -> list[PluginConfig]:
        """
        Collect all auth configs from the provider that have credentials.

        :return: List of auth plugin configs with credentials.
        """
        return [
            getattr(self.config, auth_key)
            for auth_key in AUTH_TOPIC_KEYS
            if hasattr(self.config, auth_key)
            and credentials_in_auth(getattr(self.config, auth_key))
        ]

    def _copy_matching_credentials(
        self,
        auth_confs_with_creds: list[PluginConfig],
    ) -> None:
        """
        Copy credentials from matching auth configs to the target auth config.

        :param auth_confs_with_creds: Auth configs with credentials.
        """
        for key in AUTH_TOPIC_KEYS:
            provider_auth_config = getattr(self.config, key, None)
            if provider_auth_config and not credentials_in_auth(provider_auth_config):
                for conf_with_creds in auth_confs_with_creds:
                    if conf_with_creds.matches_target_auth(provider_auth_config):
                        getattr(
                            self.config, key
                        ).credentials = conf_with_creds.credentials
                        return

    def delete_collection(self, name: str) -> None:
        """Remove a collection from this provider.

        :param name: The collection name.

        :raises UnsupportedCollection: If the collection is not found.
        """
        try:
            del self.collections_config[name]
        except KeyError:
            msg = f"Collection '{name}' not found in provider '{self.name}'."
            raise UnsupportedCollection(msg)

    def sync_collections(
        self,
        dag: EODataAccessGateway,
        strict_mode: bool,
    ) -> None:
        """
        Synchronize collections for a provider based on strict or permissive mode.

        In strict mode, removes collections not in :attr:`~eodag.api.core.EODataAccessGateway.collections_config`.
        In permissive mode, adds empty collection to config for missing types.

        :param dag: The gateway instance to use to list existing collections and to create new collection instances.
        :param strict_mode: If ``True``, remove unknown collections; if ``False``, add empty configs for them.
        """
        products_to_remove: list[str] = []
        products_to_add: list[str] = []

        for product_id in self.collections_config:
            if product_id == GENERIC_COLLECTION:
                continue

            if product_id not in dag.collections_config:
                if strict_mode:
                    products_to_remove.append(product_id)
                    continue

                empty_product = Collection.create_with_dag(
                    dag, id=product_id, title=product_id, description=NOT_AVAILABLE
                )
                dag.collections_config[product_id] = empty_product
                products_to_add.append(product_id)

        if products_to_add:
            logger.debug(
                "Collections permissive mode, %s added (provider %s)",
                ", ".join(products_to_add),
                self,
            )

        if products_to_remove:
            logger.debug(
                "Collections strict mode, ignoring %s (provider %s)",
                ", ".join(products_to_remove),
                self,
            )
            for id in products_to_remove:
                self.delete_collection(id)

    def _mm_already_built(self) -> bool:
        """Check if metadata mapping is already built (converted to querypaths/conversion)."""
        mm = getattr(self.search_config, "metadata_mapping", None)
        if not mm:
            return False

        try:
            first = next(iter(mm.values()))
        except StopIteration:
            return False

        # Consider it built if it's a tuple, or a list with second element as tuple
        if isinstance(first, tuple):
            return True
        if isinstance(first, list) and len(first) > 1 and isinstance(first[1], tuple):
            return True

        return False

    def update_from_config(self, config: Union[ProviderConfig, dict[str, Any]]) -> None:
        """Update the provider's configuration from a given config.

        :param config: The new configuration to update from.
        :raises ValidationError: If the config attempts to change the provider name.
        """
        # Prevent name changes to maintain provider identity
        source = config if isinstance(config, dict) else config.__dict__
        if (new_name := source.get("name")) and new_name != self._name:
            raise ValidationError(
                f"Cannot change provider name from '{self._name}' to '{new_name}'. "
                "Provider names are immutable after creation."
            )

        # check if metadata mapping is already built for that provider
        # this happens when the provider search plugin has already been used
        search_key = "search" if "search" in self.config else "api"
        new_conf_search = source.get(search_key, {}) or {}

        if "metadata_mapping" in new_conf_search and self._mm_already_built():
            mtd_cfg_as_conversion_and_querypath(
                deepcopy(new_conf_search["metadata_mapping"]),
                new_conf_search["metadata_mapping"],
            )

        self.config.update(config)


class ProvidersDict(UserDict[str, Provider]):
    """
    A dictionary-like collection of :class:`~eodag.api.provider.Provider` objects, keyed by provider name.

    :param providers: Initial providers to populate the dictionary.
    """

    def __contains__(self, item: object) -> bool:
        """
        Check if a provider is in the dictionary by name or :class:`~eodag.api.provider.Provider` instance.

        :param item: Provider name or Provider instance to check.
        :return: True if the provider is in the dictionary, False otherwise.
        """
        if isinstance(item, Provider):
            return item.name in self.data
        return item in self.data

    def __setitem__(self, key: str, value: Provider) -> None:
        """
        Add a :class:`~eodag.api.provider.Provider` to the dictionary.

        :param key: The name of the provider.
        :param value: The Provider instance to add.
        :raises ValueError: If the provider key already exists.
        """
        if key in self.data:
            msg = f"Provider '{key}' already exists."
            raise ValueError(msg)
        super().__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        """
        Delete a provider by name.

        :param key: The name of the provider to delete.
        :raises UnsupportedProvider: If the provider key is not found.
        """
        if key not in self.data:
            msg = f"Provider '{key}' not found."
            raise UnsupportedProvider(msg)
        super().__delitem__(key)

    def __repr__(self) -> str:
        """
        String representation of :class:`~eodag.api.provider.ProvidersDict`.

        :return: String listing provider names.
        """
        return f"ProvidersDict({list(self.data.keys())})"

    def _repr_html_(self, embeded=False) -> str:
        """
        HTML representation for Jupyter/IPython display.

        :return: HTML string representation of the :class:`~eodag.api.provider.ProvidersDict`.
        """
        longest_name = max([len(k) for k in self.keys()])
        thead = (
            f"""<thead><tr><td style='text-align: left; color: grey;'>
                {type(self).__name__}&ensp;({len(self)})
                </td></tr></thead>
            """
            if not embeded
            else ""
        )
        tr_style = "style='background-color: transparent;'" if embeded else ""
        return (
            f"<table>{thead}"
            + "".join(
                [
                    f"""<tr {tr_style}><td style='text-align: left;'>
                <details><summary style='color: grey;'>
                    <span style='color: black; font-family: monospace;'>{k}:{'&nbsp;' * (longest_name - len(k))}</span>
                    Provider(
                        {"'priority': '<span style='color: black'>" + str(v.priority) + "</span>',&ensp;"
                         if v.priority is not None else ""}
                        {"'title': '<span style='color: black'>"
                         + shorten(v.title, width=70, placeholder="[...]") + "</span>',&ensp;"
                         if v.title else ""}
                        {"'url': '" + str_as_href(v.url) + "'" if v.url else ""}
                    )
                </summary>
                    {v._repr_html_(embedded=True)}
                </details>
                </td></tr>
                """
                    for k, v in self.items()
                ]
            )
            + "</table>"
        )

    @property
    def names(self) -> list[str]:
        """
        List of provider names.

        :return: List of provider names.
        """
        return [provider.name for provider in self.data.values()]

    @property
    def groups(self) -> list[str]:
        """
        List of provider groups if exist or names.

        :return: List of provider groups if exist or names.
        """
        return list(
            set(provider.group or provider.name for provider in self.data.values())
        )

    @property
    def configs(self) -> dict[str, ProviderConfig]:
        """
        Dictionary of provider configs keyed by provider name.

        :return: Dictionary mapping provider name to :class:`~eodag.api.provider.ProviderConfig`.
        """
        return {provider.name: provider.config for provider in self.data.values()}

    @property
    def priorities(self) -> dict[str, int]:
        """
        Dictionary of provider priorities keyed by provider name.

        :return: Dictionary mapping provider name to priority integer.
        """
        return {
            provider.name: provider.config.priority for provider in self.data.values()
        }

    def get_config(self, provider: str) -> Optional[ProviderConfig]:
        """
        Get a :class:`~eodag.api.provider.ProviderConfig` from provider name.

        :param provider: The provider name.
        :return: The :class:`~eodag.api.provider.ProviderConfig` if found, otherwise None.
        """
        prov = self.get(provider)
        return prov.config if prov else None

    def filter(self, q: Optional[str] = None) -> ProvidersDict:
        """
        Return providers whose name, group, description, URL or collection matches the free-text query.

        Supports logical operators with parenthesis (``AND``/``OR``/``NOT``), quoted phrases (``"exact phrase"``),
        ``*`` and ``?`` wildcards.

        If no query is provided, returns all providers.

        :param q: Free-text parameter to filter providers. If None, returns all providers.
        :return: matching Provider objects in a :class:`~eodag.api.provider.ProvidersDict`.

        Example
        -------

        >>> from eodag.api.provider import ProvidersDict, Provider
        >>> providers = ProvidersDict()
        >>> providers['test1'] = Provider({
        ...     'name': 'test1',
        ...     'description': 'Satellite data',
        ...     'search': {'type': 'StacSearch'}
        ... })
        >>> providers['test2'] = Provider({
        ...     'name': 'test2',
        ...     'description': 'Weather data',
        ...     'search': {'type': 'StacSearch'}
        ... })
        >>> # Filter by description content
        >>> providers.filter('Satellite')
        ProvidersDict(['test1'])
        >>> # Filter with logical operators
        >>> providers['test3'] = Provider({
        ...     'name': 'test3',
        ...     'description': 'Satellite weather data',
        ...     'search': {'type': 'StacSearch'}
        ... })
        >>> providers.filter('Satellite AND weather')
        ProvidersDict(['test3'])
        >>> # Get all providers when no filter
        >>> len(providers.filter())
        3
        """
        if not q:
            # yield from self.data.values()
            return self

        free_text_query = compile_free_text_query(q)
        searchable_attributes = {"name", "group", "description", "products"}

        filtered = ProvidersDict()
        for p in self.data.values():
            searchables = {
                k: v for k, v in p.config.__dict__.items() if k in searchable_attributes
            }
            if free_text_query(searchables):
                # yield p
                filtered[p.name] = p
        return filtered

    def filter_by_name_or_group(
        self, name_or_group: Optional[str] = None
    ) -> Iterator[Provider]:
        """
        Yield providers whose name or group matches the given ``name_or_group``.

        If ``name_or_group`` is ``None``, yields all providers.

        :param name_or_group: The provider name or group to filter by. If None, yields all providers.
        :return: Iterator of matching :class:`~eodag.api.provider.Provider` objects.

        Example
        -------

        >>> from eodag.api.provider import ProvidersDict, Provider
        >>> providers = ProvidersDict()
        >>> providers['sentinel'] = Provider({'name': 'sentinel', 'group': 'esa', 'search': {'type': 'StacSearch'}})
        >>> providers['landsat'] = Provider({'name': 'landsat', 'group': 'usgs', 'search': {'type': 'StacSearch'}})
        >>> providers['modis'] = Provider({'name': 'modis', 'group': 'nasa', 'search': {'type': 'StacSearch'}})
        >>>
        >>> # Filter by exact provider name
        >>> list(p.name for p in providers.filter_by_name_or_group('sentinel'))
        ['sentinel']
        >>>
        >>> # Filter by group (case-insensitive)
        >>> list(p.name for p in providers.filter_by_name_or_group('ESA'))
        ['sentinel']
        >>>
        >>> # Get all providers when no filter
        >>> len(list(providers.filter_by_name_or_group()))
        3
        """
        if name_or_group is None:
            yield from self.data.values()
            return

        name_or_group_lower = name_or_group.lower()
        for provider in self.data.values():
            if provider.name.lower() == name_or_group_lower or (
                provider.group and provider.group.lower() == name_or_group_lower
            ):
                yield provider

    def delete_collection(self, provider: str, collection: str) -> None:
        """
        Delete a collection from a provider.

        :param provider: The provider's name.
        :param product_ID: The collection to delete.
        :raises UnsupportedProvider: If the provider or product is not found.
        """
        if provider_obj := self.get(provider):
            if collection in provider_obj.collections_config:
                provider_obj.delete_collection(collection)
            else:
                msg = f"Collection '{collection}' not found for provider '{provider}'."
                raise UnsupportedCollection(msg)
        else:
            msg = f"Provider '{provider}' not found."
            raise UnsupportedProvider(msg)

    def _share_credentials(self) -> None:
        """
        Share credentials between plugins with matching criteria
        across all providers in this dictionary.
        """
        auth_confs_with_creds: list[PluginConfig] = []
        for provider in self.values():
            auth_confs_with_creds.extend(provider._get_auth_confs_with_credentials())

        if not auth_confs_with_creds:
            return

        for provider in self.values():
            provider._copy_matching_credentials(auth_confs_with_creds)

    @staticmethod
    def _get_whitelisted_configs(
        configs: Mapping[str, Union[ProviderConfig, dict[str, Any]]],
    ) -> Mapping[str, Union[ProviderConfig, dict[str, Any]]]:
        """
        Filter configs according to the EODAG_PROVIDERS_WHITELIST environment variable, if set.

        :param configs: The dictionary of provider configurations.
        :return: Filtered configurations.
        """
        whitelist = set(os.getenv("EODAG_PROVIDERS_WHITELIST", "").split(","))
        if not whitelist or whitelist == {""}:
            return configs
        return {name: conf for name, conf in configs.items() if name in whitelist}

    def update_from_configs(
        self,
        configs: Mapping[str, Union[ProviderConfig, dict[str, Any]]],
    ) -> None:
        """
        Update providers from a dictionary of configurations.

        :param configs: A dictionary mapping provider names to configurations.
        """
        configs = self._get_whitelisted_configs(configs)
        for name, conf in configs.items():
            if isinstance(conf, dict) and conf.get("name") != name:
                if "name" in conf:
                    logger.debug(
                        "%s: config name '%s' overridden by dict key",
                        name,
                        conf["name"],
                    )
                conf = {**conf, "name": name}
            elif isinstance(conf, ProviderConfig) and conf.name != name:
                raise ValidationError(
                    f"ProviderConfig name '{conf.name}' must match dict key '{name}'"
                )

            try:
                if name in self.data:
                    self.data[name].update_from_config(conf)
                else:
                    self.data[name] = Provider(conf)

                self.data[name].collections_fetched = False

            except Exception:
                if name in self.data:
                    logger.warning(
                        "%s: skipped updating provider due to invalid config", name
                    )
                else:
                    logger.warning(
                        "%s: could not create provider from scratch using config", name
                    )
                logger.debug("Traceback:\n%s", traceback.format_exc())

        self._share_credentials()

    def update_from_config_file(self, file_path: str) -> None:
        """
        Override provider configurations with values loaded from a YAML file.

        :param file_path: The path to the configuration file.
        :raises yaml.parser.ParserError: If the YAML file cannot be parsed.
        """
        logger.info("Loading user configuration from: %s", os.path.abspath(file_path))
        with open(os.path.abspath(os.path.realpath(file_path)), "r") as fh:
            try:
                config_in_file = yaml.safe_load(fh)
                if config_in_file is None:
                    return
            except yaml.parser.ParserError as e:
                logger.error("Unable to load configuration file %s", file_path)
                raise e

        self.update_from_configs(config_in_file)

    def update_from_env(self) -> None:
        """
        Override provider configurations with environment variables values.

        Environment variables must start with ``EODAG__`` and follow a nested key
        pattern separated by double underscores ``__``.
        """

        def build_mapping_from_env(
            env_var: str, env_value: str, mapping: dict[str, Any]
        ) -> None:
            """
            Recursively build a dictionary from an environment variable.

            The environment variable must respect the pattern: ``KEY1__KEY2__[...]__KEYN``.
            It will be transformed into a nested dictionary.

            :param env_var: The environment variable key (nested keys separated by ``__``).
            :param env_value: The value from environment variable.
            :param mapping: The dictionary where the nested mapping is built.
            """
            parts = env_var.split("__")
            iter_parts = iter(parts)
            env_type = get_type_hints(PluginConfig).get(next(iter_parts, ""), str)
            child_env_type = (
                get_type_hints(env_type).get(next(iter_parts, ""))
                if isclass(env_type)
                else None
            )
            if len(parts) == 2 and child_env_type:
                try:
                    env_value = cast_scalar_value(env_value, child_env_type)
                except TypeError:
                    logger.warning(
                        f"Could not convert {parts} value {env_value} to {child_env_type}"
                    )
                mapping.setdefault(parts[0], {})
                mapping[parts[0]][parts[1]] = env_value
            elif len(parts) == 1:
                try:
                    env_value = cast_scalar_value(env_value, env_type)
                except TypeError:
                    logger.warning(
                        f"Could not convert {parts[0]} value {env_value} to {env_type}"
                    )
                mapping[parts[0]] = env_value
            else:
                new_map = mapping.setdefault(parts[0], {})
                build_mapping_from_env("__".join(parts[1:]), env_value, new_map)

        logger.debug("Loading configuration from environment variables")

        mapping_from_env: dict[str, dict[str, Any]] = {}
        for env_var in os.environ:
            if env_var.startswith("EODAG__"):
                build_mapping_from_env(
                    env_var[len("EODAG__") :].lower(),  # noqa
                    os.environ[env_var],
                    mapping_from_env,
                )

        self.update_from_configs(mapping_from_env)

    @classmethod
    def from_configs(
        cls, configs: Mapping[str, Union[ProviderConfig, dict[str, Any]]]
    ) -> Self:
        """
        Build a ProvidersDict from a configuration mapping.

        :param configs: A dictionary mapping provider names to configuration dicts or
                        :class:`~eodag.api.provider.ProviderConfig` instances.
        :return: An instance of :class:`~eodag.api.provider.ProvidersDict` populated with the given configurations.
        """
        providers = cls()
        providers.update_from_configs(configs)
        return providers



# -*- coding: utf-8 -*-
# Copyright 2021, CS GROUP - France, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional

from tests import TEST_RESOURCES_PATH, EODagTestCase
from tests.context import EOProduct, Sentinel2Driver


class TestEOProductDriverSentinel2Driver(EODagTestCase):
    def setUp(self):
        super(TestEOProductDriverSentinel2Driver, self).setUp()
        self.product = EOProduct(
            self.provider, self.eoproduct_props, collection=self.collection
        )
        self.product.properties["title"] = os.path.join(
            TEST_RESOURCES_PATH,
            "products",
            "S2A_MSIL1C_20180101T105441_N0206_R051_T31TDH_20180101T124911",
        )

    def test_driver_s2_init(self):
        """The appropriate driver must have been set"""

        def make_product(
            provider: str = "fake_provider",
            properties: dict = {},
            collection: Optional[str] = None,
        ) -> EOProduct:

            if "id" not in properties:
                properties["id"] = "9deb7e78-9341-5530-8fe8-f81fd99c9f0f"

            return EOProduct(provider, properties, collection=collection)

        for test in [
            {
                "label": "Test match driver sentinel 2 by collection name",
                "product": make_product(collection="S2_MSI_L1C"),
                "match": True,
            },
            {
                "label": "Test match driver sentinel 2 by collection name (sensitive case)",
                "product": make_product(collection="s2_msi_l1c"),
                "match": True,
            },
            {
                "label": "Test mismatch driver sentinel 2 by collection name (sensitive case)",
                "product": make_product(collection="s3_msi_l1c"),
                "match": False,
            },
            {
                "label": "Test match driver sentinel 2 by property constellation",
                "product": make_product(properties={"constellation": "sentinel2"}),
                "match": True,
            },
            {
                "label": "Test match driver sentinel 2 by property constellation (sensitive case)",
                "product": make_product(properties={"constellation": "SeNtInEl2"}),
                "match": True,
            },
            {
                "label": "Test match driver sentinel 2 by property constellation (specials chars)",
                "product": make_product(
                    properties={"constellation": "--[sentinel #2]--"}
                ),
                "match": True,
            },
            {
                "label": "Test mismatch driver sentinel 2 by property constellation (specials chars)",
                "product": make_product(properties={"constellation": "sentinel1"}),
                "match": False,
            },
            {
                "label": "Test match driver sentinel 2 by property platform",
                "product": make_product(properties={"platform": "S2A"}),
                "match": True,
            },
            {
                "label": "Test match driver sentinel 2 by property platform (sensitive case)",
                "product": make_product(properties={"platform": "s2b"}),
                "match": True,
            },
            {
                "label": "Test mismatch driver sentinel 2 by property platform",
                "product": make_product(properties={"platform": "S3X"}),
                "match": False,
            },
        ]:
            self.assertEqual(
                isinstance(test["product"].driver, Sentinel2Driver),
                test["match"],
                "Fail: {}".format(test["label"]),
            )

    def test_driver_s2_guess_asset_key_and_roles(self):
        """The driver must guess appropriate asset key and roles"""
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles("", self.product),
            (None, None),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "tiles/31/T/DJ/2018/1/28/0/bla_TCI.jp2", self.product
            ),
            ("TCI", ["visual"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "tiles/31/T/DJ/2018/1/28/0/B01.jp2", self.product
            ),
            ("B01", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "http://foo/1/28/0/bla_bla-B02_10m.tif", self.product
            ),
            ("B02", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/bla_bla-B02_20m.tiff", self.product
            ),
            ("B02_20m", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/MSK_bla_B02.tiff", self.product
            ),
            ("MSK_bla_B02", ["data-mask"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/Foo-bar.xml", self.product
            ),
            ("Foo-bar.xml", ["metadata"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/tilejson.json?foo", self.product
            ),
            ("tilejson.json", ["metadata"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/thumbnail.png", self.product
            ),
            ("thumbnail.png", ["thumbnail"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/Foo_bar-ql.jpg", self.product
            ),
            ("ql.jpg", ["overview"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/foo.bar?baz", self.product
            ),
            ("foo.bar", ["auxiliary"]),
        )
        # no extension: return None to keep original key and roles
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/some-key", self.product
            ),
            (None, None),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "https://example.com/path/no-ext-file?foo=bar", self.product
            ),
            (None, None),
        )
